#!/usr/bin/env python
# ===================================================================
# ScriptName:   gui_tgt
# Purpose:      User interface for selecting SPACEtomo targets
#               More information at http://github.com/eisfabian/SPACEtomo
# Author:       Fabian Eisenstein
# Created:      2024/03/20
# Revision:     v1.2
# Last Change:  2024/11/27: added reacquire button
#               2024/10/02: replaced text inputs with sliders for targeting settings
#               2024/09/25: added check to save acquisition settings
#               2024/09/04: small quality of life improvements after feedback
#               2024/08/22: adjust tilt axis position to tracking target
#               2024/08/20: fixed tilt axis plot
#               2024/08/09: Sped up loading by 30-40% using toNumpy, float32 and tiled textures, added menu to change target area of point
#               2024/08/08: all old functions reimplemented, added option to make target tracking target
#               2024/08/07: branched out tgt.py for Targets class, added adding and deleting targets
#               2024/07/31: rewrote map, segmentation and target loading
#               2024/07/30: started rewrite using gui module
#               2024/04/09: fixes after Rado Krios test, added default binning to speed up loading
#               2024/04/02: fixes after Krios 1 test, allowed click + D to delete and click + G to add geo point, added tilt axis
#               2024/03/27: added tooltips, added ask for save on close, added variable tooltip for targets on plot
#               2024/03/26: added geo points handling
#               2024/03/25: added dir as argument to be called from SPACEtomo script, added thicker outlines for targets, added None class to not show any overlay, added generic thumbnail, minor text fixes
#               2024/03/22: add table for map selection window
#               2024/03/21: Copied all manipulation functions and considered proper pixel size, added thumbnail creation and map selection window
#               2024/03/20: Copy most if inspection tab from SPACEtomo_TI
# ===================================================================

import os
os.environ["__GLVND_DISALLOW_PATCHING"] = "1"           # helps to minimize Segmentation fault crashes on Linux when deleting textures
import sys
import json
import time
from pathlib import Path
try:
    import dearpygui.dearpygui as dpg
except:
    print("ERROR: DearPyGUI module not installed! If you cannot install it, please run the GUI from an external machine.")
    sys.exit()
from PIL import Image, ImageDraw
Image.MAX_IMAGE_PIXELS = None
import numpy as np

from SPACEtomo.modules import ext as space_ext
from SPACEtomo.modules.gui import gui
from SPACEtomo.modules.gui.thmb import MapWindow
from SPACEtomo.modules.gui.flm import FlmWindow
from SPACEtomo.modules.gui.menu import Menu
from SPACEtomo.modules.gui.plot import Plot
from SPACEtomo.modules.gui.map import MMap, Segmentation
from SPACEtomo.modules.tgt import Targets
from SPACEtomo.modules import utils
from SPACEtomo.modules.utils import log
from SPACEtomo import __version__

import faulthandler
faulthandler.enable()

class TargetGUI:

    def mouseClick(self, sender, app_data):
        """Handle mouse clicks when plot is hovered within map boundaries."""

        # Get mouse coords in plot coord system
        mouse_coords = np.array(dpg.get_plot_mouse_pos())

        # Delegate to sub-window if open
        if dpg.is_item_shown(self.lm_window.window):
            self.lm_window.mouseClick(mouse_coords)
            return

        # Check if click needs to be processed
        if not dpg.is_item_hovered(self.plot.plot) or not self.plot.withinBounds(mouse_coords) or self.inspected:
            return
        
        # Convert plot coords to image coords in px
        img_coords = self.loaded_map.microns2px(mouse_coords)

        # Left mouse button functions
        if dpg.is_mouse_button_down(dpg.mvMouseButton_Left) and (dpg.is_key_down(dpg.mvKey_LShift) or dpg.is_key_down(dpg.mvKey_RShift)):
            if self.targets.addTarget(img_coords, new_area=dpg.is_key_down(dpg.mvKey_T)):
                self.showTargets()
                dpg.show_item(self.menu_right.all_elements["butsave"])

        # Right mouse button functions
        elif dpg.is_mouse_button_down(dpg.mvMouseButton_Right) or (dpg.is_mouse_button_down(dpg.mvMouseButton_Left) and dpg.is_key_down(dpg.mvKey_E)):
            self.showTargetMenu(img_coords)  

        # Middle mouse button functions
        elif dpg.is_mouse_button_down(dpg.mvMouseButton_Middle) or (dpg.is_mouse_button_down(dpg.mvMouseButton_Left) and dpg.is_key_down(dpg.mvKey_G)):
            if self.targets.addGeoPoint(img_coords):
                self.showTargets()
                dpg.show_item(self.menu_right.all_elements["butsave"])

    def mouseRelease(self, sender, app_data):
        """Handle mouse release and check if any drag points were moved."""

        # Delegate to sub-window if open
        if dpg.is_item_shown(self.lm_window.window):
            self.lm_window.mouseRelease()
            return

        # Only process when targets are editable
        if not self.targets or self.inspected:
            return

        # Check all drag points for changes
        update = False
        for point in self.plot.drag_points:
            update = update or self.checkDragPoint(point["plot"])

        if update:
            self.plot.clearSeries(["temp_drag"])
            self.showTargets()
            dpg.show_item(self.menu_right.all_elements["butsave"])

    def dragPointUpdate(self, sender, app_data, user_data):
        """Handles dragging of points and updates scatter point to mouse position."""

        # Don't allow changes on inspected targets
        if self.inspected: return

        # Update scatter indicator for drag point (drag points are always plotted behind images)
        coords = dpg.get_value(sender)[:2]
        self.plot.clearSeries(["temp_drag"])
        self.plot.addSeries([coords[0]], [coords[1]], label="temp_drag", theme="drag_scatter_theme")

    def checkDragPoint(self, point):
        """Compares drag points to targets coords and updates targets if they have changed."""

        # Get coords from drag point value
        coords = np.array(dpg.get_value(point)[:2])
        
        # Get area and point IDs from user data embedded in drag point
        user_data = dpg.get_item_user_data(point).split("_")
        point_type = user_data[0]
        point_id = np.array(user_data[1:], dtype=int)
        
        # Transform points to plot points for comparison
        if point_type == "pt":
            old_coords = self.loaded_map.px2microns(self.targets.areas[point_id[0]].points[point_id[1]])
        elif point_type == "geo":
            old_coords = self.loaded_map.px2microns(self.targets.areas[point_id[0]].geo_points[point_id[1]])
        else:
            log(f"ERROR: Invalid drag point! [{user_data}]")
            return
        
        # Go to next points if coords have not changed
        if np.all(coords == old_coords):
            return False
        else:
            # Get camera dims in microns
            rec_dims = np.array(self.tgt_params.weight.shape) * self.loaded_map.pix_size / 10000

            # Clip coords to map size
            coords[0] = np.clip(coords[0], rec_dims[1] / 2, self.loaded_map.dims_microns[0] - rec_dims[1] / 2)
            coords[1] = np.clip(coords[1], rec_dims[0] / 2, self.loaded_map.dims_microns[1] - rec_dims[0] / 2)

            # Update coords if they have changed
            if point_type == "pt":
                self.targets.areas[point_id[0]].points[point_id[1]] = self.loaded_map.microns2px(coords)
            elif point_type =="geo":
                self.targets.areas[point_id[0]].geo_points[point_id[1]] = self.loaded_map.microns2px(coords)

            return True

    def updateMapList(self, list_only=False):
        """Checks for new maps in folder."""

        prev_map_num = len(self.map_list)
        prev_tgt_num = sum(self.map_list_tgtnum)

        # Find all maps with segmentation (and existing map, which sometimes is saved later due to threading)
        seg_list = sorted(self.cur_dir.glob("*_seg.png"))
        self.map_list = [seg.name.split("_seg.png")[0] for seg in seg_list if (seg.parent / (seg.name.split("_seg.png")[0] + ".png")).exists()]

        if self.map_list and not list_only:        # distinction necessary for creating map list before GUI is set up
            # Update combo menu
            dpg.configure_item(self.menu_left.all_elements["map"], items=self.map_list, label=f"({len(self.map_list)} maps)")
            # Set value to loaded map
            if self.map_name:
                dpg.set_value(self.menu_left.all_elements["map"], self.map_name)
            else:
                dpg.set_value(self.menu_left.all_elements["map"], self.map_list[0])

        # Find selected targets per lamella
        self.map_list_tgtnum = []
        for m_name in self.map_list:
            point_files = sorted(self.cur_dir.glob(m_name + "_points*.json"))
            point_num = 0
            if len(point_files) > 0:
                for file in point_files:
                    # Load json data
                    with open(file, "r") as f:
                        point_data = json.load(f, object_hook=utils.revertArray)
                        point_num += len(point_data["points"])
            self.map_list_tgtnum.append(point_num)

        # Check for changes
        if len(self.map_list) != prev_map_num or sum(self.map_list_tgtnum) != prev_tgt_num:
            return True
        else:
            return False
        
    def checkPointFiles(self):
        """Checks if new point file of loaded map was created."""

        # Don't check if no map is loaded or targets are already loaded (or selected)
        if not self.map_name or (self.targets and len(self.targets) > 0) or self.checked_point_files:
            return
        
        # Get all point files for lamella map
        point_files = sorted(self.cur_dir.glob(self.map_name + "_points*.json"))
        if len(point_files) > 0:
            # Check if any points in point files
            point_num = 0
            for file in point_files:
                # Load json data
                with open(file, "r") as f:
                    point_data = json.load(f, object_hook=utils.revertArray)
                    point_num += len(point_data["points"])

            # Arrange opening of prompt to load point files
            if point_num > 0:
                log(f"NOTE: Target coordinates for the loaded map were found!")
                # Need to open info box on next mouse move because info box requires split_frame function which can't be called from within main loop
                with dpg.handler_registry(tag="check_point_file"):
                    dpg.add_mouse_move_handler(callback=self.confirmLoadPointFiles)

    def confirmLoadPointFiles(self, sender=None, app_data=None, user_data=None):
        """Loads point file after user confirmation."""

        # Reset mouse move handler
        if dpg.does_item_exist("check_point_file"):
            dpg.delete_item("check_point_file")

        # Open box unless already called by box
        if not user_data:
            gui.showInfoBox("New target coordinates", "Target coordinates for the currently loaded map were found. Do you want to load them?", callback=self.confirmLoadPointFiles, options=["Load", "Cancel"], options_data=[True, False])
            self.checked_point_files = True
            return

        # Check for info box input
        if user_data and dpg.does_item_exist(user_data[0]):
            dpg.delete_item(user_data[0])
            dpg.split_frame()

        # Check user choice
        if user_data[1]:
            self.loadTargetsFile()

    def selectMap(self, sender=None, app_data=None, user_data=None, next_map=False):
        """Handles map selection from several buttons."""

        # Check for valid user_data
        if user_data and user_data in self.map_list:
            self.map_name = user_data
        else:            
            self.map_name = dpg.get_value(self.menu_left.all_elements["map"])

        if next_map:
            map_id = self.map_list.index(self.map_name) + 1
            if map_id >= len(self.map_list):
                map_id = 0
                log("WARNING: Reached end of list of maps. Loading first map!")
            self.map_name = self.map_list[map_id]

        dpg.set_value(self.menu_left.all_elements["map"], self.map_name)

        # Close map window if open
        self.map_window.hide()

        # Update map list on every button click
        if self.updateMapList():
            # Update fill map table
            self.map_window.update(self.map_name, self.map_list, self.map_list_tgtnum)
        else:
            # Update only selected map
            self.map_window.updateMap(self.map_name)

        self.loadMap()

    def loadMap(self):
        """Loads map from file and all associated data."""

        file_path = self.cur_dir.joinpath(self.map_name + ".png")

        # Update GUI
        self.menu_left.hide()
        self.menu_right.hide()
        self.plot.updateLabel(file_path.name + " loading...")
        self.status.update("Loading map...", box=True)

        # Reset FLM window plots
        self.lm_window.clearAll()
        # Reset plot
        self.plot.clearAll()
        # Clear out texture references because textures have been deleted when plot was cleared
        self.target_overlays = {}
        self.makeTargetOverlay() # Pre-generate overlays to remove lag when selecting first target

        # Load map file
        self.loaded_map = MMap(file_path, pix_size=self.model.pix_size * 10, status=self.status)
        self.loaded_map.checkBinning(default_binning=dpg.get_value(self.menu_left.all_elements["inpbin"]))
        self.status.update("Plotting map...", box=True)
        #map_texture = self.loaded_map.getTexture()
        map_textures, bounds = self.loaded_map.getTileTextures()

        # Plot map
        start = time.time()
        #self.plot.addImg(map_texture, [[0, self.loaded_map.dims_microns[0]], [0, self.loaded_map.dims_microns[1]]], self.loaded_map.binning, label="map")
        for i, (tex, bound) in enumerate(zip(map_textures, bounds)):
            self.plot.addImg(tex, bound, self.loaded_map.binning, label=f"map{i}")
        log(f"Map was plotted in {time.time() - start} s.")
        self.plot.resetZoom()
        self.plot.updateLabel(self.map_name)

        self.lm_window.loadEM(self.loaded_map, map_textures, bounds)

        # Load Segmentation
        self.status.update("Loading segmentation...", box=True)
        self.segmentation = Segmentation(self.loaded_map.file.parent / (self.loaded_map.file.stem + "_seg.png"), self.model, self.loaded_map.pix_size)

        # Show tilt axis
        self.showTiltAxis()

        # Load targets
        self.status.update("Loading targets...", box=True)
        self.targets = Targets(map_dir=self.cur_dir, map_name=self.map_name, map_dims=self.loaded_map.img.shape, tgt_params=self.tgt_params, map_pix_size=self.loaded_map.pix_size)

        self.loadTargetsFile()

        # Reset checked for point files status
        self.checked_point_files = False

        # Update GUI
        #self.menu_left.unlockRows(["filters"])     # Not yet implemented
        self.menu_left.unlockRows(["reacquire"])
        if self.segmentation.valid:         # Only show class selection if segmentation was performed
            self.menu_left.unlockRows(["class_list", "class_buttons"])
            self.menu_right.unlockRows(["settings1", "settings2", "settings3", "settings4"])

        self.menu_left.show()

        self.status.update()

    def loadTargetsFile(self):
        """Loads targets from file and updates GUI."""

        self.targets.loadAreas()
        self.showTargets()

        # Fill settings input fields from loaded targets
        self.setAcquisitionSettings()

        # Get class scores for targets
        if self.segmentation.valid and self.targets and len(self.tgt_params.target_list) > 1:
            self.getTargetClassScores()

        # Check inspected
        if (self.cur_dir / (self.map_name + "_inspected.txt")).exists():
            self.inspected = True
            dpg.show_item("inspected")
        else:
            self.inspected = False
            dpg.hide_item("inspected")

        # Update GUI
        if not self.inspected:
            self.showTargetAreaButtons()
            self.menu_right.unlockRows(["areas", "delete", "acquisition1", "acquisition2", "acquisition3", "acquisition4", "acquisition5", "acquisition6", "buttons1", "buttons2"])
            self.menu_right.show()

            # Reset save button
            dpg.hide_item(self.menu_right.all_elements["butsave"])

    def loadOverlay(self, *args, cats=[]):
        """Plots overlay of selected categories."""

        # Update GUI
        self.menu_left.hide()
        self.menu_right.hide()
        self.status.update("Loading overlay...", box=True)

        if cats:
            selected_cats = cats
        else:
            # Get selected classes from checkboxes
            selected_cats = self.getClassSelection(self.menu_left)

        # Check if selected classes have changed
        if not all([cat in self.overlay for cat in selected_cats]):

            # Clear previous overlay
            self.plot.clearImg(self.plot.getImgByKeyword("seg") + [""])     # needs empty list entry, otherwise all items will be cleared in case of no hits

            # If no classes selected, don't load new overlay
            if selected_cats:

                # Get texture
                #mask_texture = self.segmentation.getMaskTexture(selected_cats)
                mask_textures, bounds = self.segmentation.getMaskTileTextures(selected_cats)

                # Plot map
                #start = time.time()
                #self.plot.addImg(mask_texture, [[0, self.loaded_map.dims_microns[0]], [0, self.loaded_map.dims_microns[1]]], self.segmentation.binning, label="seg")
                #log(f"Overlay plotted in {time.time() - start} s.")

                # Plot map
                start = time.time()
                for i, (tex, bound) in enumerate(zip(mask_textures, bounds)):
                    self.plot.addImg(tex, bound, self.loaded_map.binning, label=f"seg{i}")
                log(f"Overlay plotted in {time.time() - start} s.")

            # Update overlay classes
            self.overlay = selected_cats

        # Update GUI
        self.menu_left.show()
        if not self.inspected:
            self.menu_right.show()
        self.status.update()

    def showTiltAxis(self, point=None):
        """Plots tilt axis."""

        # Clear previous tilt axis
        self.plot.clearSeries(["tilt_axis"])

        # Get tilt axis rotation matrix
        theta = np.radians(self.mic_params.view_ta_rotation)
        rotM = np.array([[np.cos(theta), -np.sin(theta)], [np.sin(theta), np.cos(theta)]])
        axis_unit = rotM @ np.array([1, 0])

        if point is None:
            # Get centered tilt axis
            axis_length = min(abs(self.loaded_map.img.shape[0] / axis_unit[0]), abs(self.loaded_map.img.shape[1] / axis_unit[1]))               # total axis length
            axis_length1 = axis_length2 = axis_length / 2
            offset = np.array([self.loaded_map.img.shape[0] / 2, self.loaded_map.img.shape[1] / 2])                                             # Offset to center of map
        else:
            # Get tilt axis centered on point
            axis_length1 = min(abs(point[0] / axis_unit[0]), abs(point[1] / axis_unit[1]))                                                      # Split axis in lengths from point to edges of map
            axis_length2 = min(abs((self.loaded_map.img.shape[0] - point[0]) / axis_unit[0]), abs((self.loaded_map.img.shape[1] - point[1]) / axis_unit[1]))
            if abs(axis_unit[0]) > abs(axis_unit[1]):
                axis_length1 = -axis_length1
                axis_length2 = -axis_length2
            offset = np.array([self.loaded_map.img.shape[0] - self.targets.areas[0].points[0][0], self.targets.areas[0].points[0][1]])          # Apply offset coords from point considering axis inversion
        axis_points = (np.array([rotM @ np.array([-axis_length1, 0]), rotM @ np.array([axis_length2, 0])]) + offset) * self.loaded_map.pix_size / 10000

        # Add custom series to plot.series list
        self.plot.addCustomPlot(self.plot.series, label="tilt_axis", plot=dpg.add_line_series(list(axis_points[:, 1]), list(axis_points[:, 0]), parent=self.plot.x_axis), theme="axis_theme")

    def filterMap(self):
        """Loads map filtered with LISC."""

        # TODO
        log("Filter map")

    def reacquireMap(self):
        """Marks map to be reacquired by changing file names."""

        mrc_file = self.cur_dir.parent / (self.map_name + ".mrc")
        if mrc_file.exists():
            counter = 0
            while (new_mrc_file := self.cur_dir.parent / (self.map_name + f"_old{counter}.mrc")).exists():
                counter += 1
            new_map_name = self.map_name + f"_old{counter}"
            # Rename mrc file and all generated files
            mrc_file.replace(new_mrc_file)
            if (self.cur_dir / (self.map_name + ".png")).exists():
                (self.cur_dir / (self.map_name + ".png")).replace(self.cur_dir / (new_map_name + ".png"))
            if (self.cur_dir / (self.map_name + "_seg.png")).exists():
                (self.cur_dir / (self.map_name + "_seg.png")).replace(self.cur_dir / (new_map_name + "_seg.png"))
            if (self.cur_dir / "thumbnails" / (self.map_name + ".png")).exists():
                (self.cur_dir / "thumbnails" / (self.map_name + ".png")).replace(self.cur_dir / "thumbnails" / (new_map_name + ".png"))
            # Provide empty point file
            if not list(self.cur_dir.glob(new_map_name + "_points*.json")):
                Targets(map_dir=self.cur_dir, map_name=new_map_name, map_dims=self.loaded_map.img.shape, tgt_params=self.tgt_params, map_pix_size=self.loaded_map.pix_size).exportTargets()
            # Mark new name as inspected and original name (to not stall current run)
            (self.cur_dir / (self.map_name + "_inspected.txt")).touch(exist_ok=True)
            (self.cur_dir / (new_map_name + "_inspected.txt")).touch(exist_ok=True)

            self.map_name = new_map_name

            log(f"DEBUG: Reacquire map {mrc_file} on next SPACEtomo run!")
            gui.showInfoBox("INFO", "This map will be reacquired on the next SPACEtomo run. If you want to reacquire now, please stop the SPACEtomo run and start it again.")
        else:
            log(f"ERROR: Cannot find existing map MRC file. Reacquisition could not be scheduled.")
            gui.showInfoBox("ERROR", "Cannot find existing map MRC file. Reacquisition could not be scheduled!")

    def applyClassSelection(self, sender, app_data, user_data):
        """Applies checked classes to target list."""

        # Check for user data
        if user_data:
            menu, element = user_data

        # Set target_list as default
        if not element:
            element = self.menu_right.all_elements["target_list"]

        # Get selected classes from checkboxes
        selected_cats = self.getClassSelection(menu)

        # Fill input
        dpg.set_value(element, ",".join(selected_cats))


    def getClassSelection(self, menu):
        """Gets selected classes from checkboxes."""

        selected_cats = []
        for cat in self.model.categories:
            if dpg.get_value(menu.all_elements[cat]):
                selected_cats.append(cat)

        return selected_cats

    def showTargets(self):
        """Updates target overlays on plot."""

        # Skip rest if currently no targets
        if not self.targets:
            return
        
        # Clear targets from plot
        self.clearTargets(plot_only=True)
        self.plot.clearSeries(["temp_drag"])

        # Generate target overlay texture
        if not self.target_overlays or not dpg.does_item_exist(self.target_overlays["target"]):
            self.makeTargetOverlay()

        # Get map dims
        dims = np.flip(np.array(self.loaded_map.img.shape))

        tgt_counter = 0
        geo_counter = 0
        for t, target_area in enumerate(self.targets.areas):
            if len(target_area.points) == 0: continue
            # Transform coords to plot
            points = np.array([self.loaded_map.px2microns(point) for point in target_area.points])

            # Scatter plot of targets
            self.plot.addSeries(points[:, 0], points[:, 1], label=f"tgtscatter_{t}", theme=f"scatter_theme{t}")

            # Go over all target points
            scaled_overlay_dims = self.target_overlays["tgtdims"] * self.loaded_map.pix_size / 10000
            for p, (x, y) in enumerate(points):
                # Add draggable point
                self.plot.addDragPoint(x, y, label=f"tgt_{tgt_counter + 1}", callback=self.dragPointUpdate, user_data=f"pt_{t}_{p}", color=gui.THEME_COLORS[t % len(gui.THEME_COLORS)])
                # Show graphical overlays
                bounds = ((x - scaled_overlay_dims[1] / 2, x + scaled_overlay_dims[1] / 2), (y - scaled_overlay_dims[0] / 2, y + scaled_overlay_dims[0] / 2))
                if p == 0:
                    self.plot.addOverlay(self.target_overlays["track"], bounds, label=f"tgt_{t}_{p}")
                else:
                    self.plot.addOverlay(self.target_overlays["target"], bounds, label=f"tgt_{t}_{p}")

                tgt_counter += 1

            if tgt_counter > 0 and len(target_area.geo_points) > 0:
                # Transform geo coords to plot
                points = np.array([self.loaded_map.px2microns(point) for point in target_area.geo_points])

                # Scatter plot of geo points
                self.plot.addSeries(points[:, 0], points[:, 1], label=f"geoscatter_{t}", theme=f"scatter_theme{t}")

                # Go over all geo points
                scaled_overlay_dims = self.target_overlays["geodims"] * self.loaded_map.pix_size / 10000
                for p, (x, y) in enumerate(points):
                    # Add draggable point
                    self.plot.addDragPoint(x, y, label=f"geo_{geo_counter}", callback=self.dragPointUpdate, user_data=f"geo_{t}_{p}", color=gui.THEME_COLORS[4])
                    # Show graphical overlays
                    bounds = ((x - scaled_overlay_dims[1] / 2, x + scaled_overlay_dims[1] / 2), (y - scaled_overlay_dims[0] / 2, y + scaled_overlay_dims[0] / 2))
                    self.plot.addOverlay(self.target_overlays["geo"], bounds, label=f"geo_{t}_{p}")

                    geo_counter += 1

        # Show tilt axis centered on tracking target
        self.showTiltAxis(self.targets.areas[0].points[0])

    def targetTooltip(self):
        """Configures plot tooltip to contain target information."""

        mouse_coords = np.array(dpg.get_plot_mouse_pos())

        # Convert plot coords to image coords in px
        if self.loaded_map and self.targets:
            img_coords = self.loaded_map.microns2px(mouse_coords)

            # Get camera dims
            rec_dims = np.array(self.tgt_params.weight.shape)

            # Check if coords are too close to existing point
            closest_point_id, in_range = self.targets.getClosestPoint(img_coords, np.min(rec_dims))

            if in_range:
                dpg.set_value("tt_heading", "Target information:")

                """
                # Set point information
                area = f"Area: {closest_point_id[0] + 1}\n" if len(self.targets.areas) > 1 else ""
                score, dist = self.targets.areas[closest_point_id[0]].getPointInfo(closest_point_id[1])
                score = round(score, 2) if score < 100 else "manual"
                dist *= self.loaded_map.pix_size / 10000
                info = f"{area}Target: {closest_point_id[1] + 1}\nScore: {score}\nIS: {round(dist, 2)} µm"
                """
                info = self.assembleTargetInfo(closest_point_id)

                dpg.set_value("tt_text", info)
                return

        dpg.set_value("tt_heading", "Target manipulation:")
        dpg.set_value("tt_text", "- Drag target to reposition\n- Shift + left click to add target\n- Right click to open target editing\n- Middle click to add geo point")

    def showTargetMenu(self, img_coords):
        """Configures target menu when right clicking on point."""

        # Get camera dims
        rec_dims = np.array(self.tgt_params.weight.shape)

        # Check if coords are too close to existing point
        closest_point_id, in_range = self.targets.getClosestPoint(img_coords, np.min(rec_dims))

        if in_range:
            dpg.set_value(self.menu_tgt.all_elements["heading_txt"], "Target information:")

            info = self.assembleTargetInfo(closest_point_id)
            """
            # Set area information if more than one target area
            area = f"Area: {closest_point_id[0] + 1}\n" if len(self.targets.areas) > 1 else ""

            # Get score and distance from tracking target
            final_score, dist = self.targets.areas[closest_point_id[0]].getPointInfo(closest_point_id[1])
            score_text = round(final_score, 2) if final_score < 100 else "manual"
            dist *= self.loaded_map.pix_size / 10000

            # Get class scores if present
            class_scores = ""
            if self.targets.areas[closest_point_id[0]].class_scores:
                for cat, score in self.targets.areas[closest_point_id[0]].class_scores.items():
                    if score[closest_point_id[1]] > 0:
                        # Format class scores to show percentage of final score
                        class_scores += f"- {cat}:{' ' * (max([len(cat_name) for cat_name in self.tgt_params.target_list]) + 1 - len(cat))}{round(score[closest_point_id[1]], 2)}\n"
            
            # Format and set info text
            info = f"{area}Target: {closest_point_id[1] + 1}\nScore: {score_text}\n{class_scores}IS: {round(dist, 2)} µm"
            """
            dpg.set_value(self.menu_tgt.all_elements["info_txt"], info)

            # Configure buttons
            dpg.configure_item(self.menu_tgt.all_elements["btn_del"], user_data=closest_point_id)
            if closest_point_id[1] > 0:
                # Only show option when target is not a tracking target
                dpg.configure_item(self.menu_tgt.all_elements["btn_trk"], user_data=closest_point_id)
                dpg.show_item(self.menu_tgt.all_elements["btn_trk"])
            else:
                dpg.hide_item(self.menu_tgt.all_elements["btn_trk"])

            # Configure area selection
            if len(self.targets.areas) > 1:
                dpg.configure_item(self.menu_tgt.all_elements["area"], items=[area + 1 for area in range(len(self.targets.areas))], default_value=closest_point_id[0] + 1, user_data=closest_point_id)
                self.menu_tgt.unlockRows(["area"])
            else:
                self.menu_tgt.lockRows(["area"])

            # Configure optimization button (only show when segmentation was loaded)
            if self.segmentation.valid:
                dpg.configure_item(self.menu_tgt.all_elements["btn_opt"], user_data=closest_point_id)
                dpg.show_item(self.menu_tgt.all_elements["btn_opt"])
            else:
                dpg.hide_item(self.menu_tgt.all_elements["btn_opt"])

            # Unlock rows
            self.menu_tgt.unlockRows(["info", "buttons"])

            # Get UI mouse coords for window placement
            mouse_coords_global = dpg.get_mouse_pos(local=False)
            dpg.set_item_pos("win_tgt", mouse_coords_global)
            dpg.show_item("win_tgt")

    def assembleTargetInfo(self, closest_point_id):
        """Creates string with target info for display."""

        # Set area information if more than one target area
        area = f"Area: {closest_point_id[0] + 1}\n" if len(self.targets.areas) > 1 else ""

        # Get score and distance from tracking target
        final_score, dist = self.targets.areas[closest_point_id[0]].getPointInfo(closest_point_id[1])
        score_text = round(final_score, 2) if final_score < 100 else "manual"
        dist *= self.loaded_map.pix_size / 10000

        # Get class scores if present (and if menu is in advanced mode)
        class_scores = ""
        if self.menu_left.show_advanced and self.targets.areas[closest_point_id[0]].class_scores:
            for cat, score in self.targets.areas[closest_point_id[0]].class_scores.items():
                if score[closest_point_id[1]] > 0:
                    # Format class scores to show percentage of final score
                    class_scores += f"- {cat}:{' ' * (max([len(cat_name) for cat_name in self.tgt_params.target_list]) + 1 - len(cat))}{round(score[closest_point_id[1]], 2)}\n"
        
        # Format and set info text
        if dist >= self.mic_params.IS_limit - 0.5:
            warning_is = "[!]"
        else:
            warning_is = ""
        return f"{area}Target: {closest_point_id[1] + 1}\nScore: {score_text}\n{class_scores}IS: {round(dist, 2)}{warning_is} µm"

    def showTargetAreaButtons(self):
        """Determines which target area actions are available."""

        # Split areas button
        if len(self.targets) > 0:
            dpg.show_item(self.menu_right.all_elements["butsplit"])
        else:
            dpg.hide_item(self.menu_right.all_elements["butsplit"])

        # Redistribute targets and merge buttons
        if len(self.targets.areas) > 1:
            dpg.show_item(self.menu_right.all_elements["butdist"])
            dpg.show_item(self.menu_right.all_elements["butmerge"])
        else:
            dpg.hide_item(self.menu_right.all_elements["butdist"])
            dpg.hide_item(self.menu_right.all_elements["butmerge"])

    def openClassSelection(self, sender, app_data, user_data):
        """Opens selection window for classes."""

        # Get current text input
        selected_cats = [value.strip() for value in dpg.get_value(app_data[1]).split(",")]

        # Adjust checkboxes accordingly
        for cat in self.model.categories:
            if cat in selected_cats:
                dpg.set_value(self.menu_sel.all_elements[cat], True)
            else:
                dpg.set_value(self.menu_sel.all_elements[cat], False)
            # Add text field id from app_data
            dpg.configure_item(self.menu_sel.all_elements[cat], user_data=[self.menu_sel, app_data[1]])

        # Get UI mouse coords for window placement
        mouse_coords_global = dpg.get_mouse_pos(local=False)
        dpg.set_item_pos("win_sel", mouse_coords_global)
        dpg.show_item("win_sel")

    def runTargetSelection(self):
        """Calls the target selection procedure."""

        # Check for existing point files and delete them
        point_files = sorted(self.cur_dir.glob(self.map_name + "_points*.json"))
        for file in point_files:
            # Delete file
            file.unlink()

        # Delete previous targets
        self.clearTargets()

        # Update tgt params from GUI settings
        self.tgt_params.target_list = [cat.strip() for cat in dpg.get_value(self.menu_right.all_elements["target_list"]).split(",")]
        self.tgt_params.penalty_list = [cat.strip() for cat in dpg.get_value(self.menu_right.all_elements["avoid_list"]).split(",")]
        self.tgt_params.parseLists(self.model)
        self.tgt_params.checkLists(self.model)

        self.tgt_params.sparse = dpg.get_value(self.menu_right.all_elements["sparse_targets"])
        self.tgt_params.edge = dpg.get_value(self.menu_right.all_elements["target_edge"])
        self.tgt_params.penalty = dpg.get_value(self.menu_right.all_elements["penalty_weight"])
        self.tgt_params.threshold = dpg.get_value(self.menu_right.all_elements["target_score_threshold"])
        self.tgt_params.max_tilt = dpg.get_value(self.menu_right.all_elements["max_tilt"])
        self.tgt_params.extra_track = dpg.get_value(self.menu_right.all_elements["extra_tracking"])

        self.mic_params.IS_limit = dpg.get_value(self.menu_right.all_elements["IS_limit"])

        # Load overlay
        self.status.update("Loading overlay...", box=True)
        self.loadOverlay(cats=self.tgt_params.target_list)

        # Update GUI
        self.menu_left.hide()
        self.menu_right.hide()

        # Run target selection
        self.status.update("Selecting targets...\n(See console output for details!)", box=True)
        space_ext.runTargetSelection(self.cur_dir, self.map_name, self.tgt_params, self.mic_params, self.model, save_final_plot=False)

        # Plot targets
        self.targets = Targets(map_dir=self.cur_dir, map_name=self.map_name, map_dims=self.loaded_map.img.shape, tgt_params=self.tgt_params, map_pix_size=self.loaded_map.pix_size)
        self.targets.loadAreas()
        self.showTargets()
        self.showTargetAreaButtons()

        # Get class specific scores
        self.getTargetClassScores()

        # Update GUI
        self.menu_left.show()
        dpg.hide_item(self.menu_right.all_elements["butsave"])     # Only activate save button when targets were changed
        if not self.inspected:
            self.menu_right.show()
        self.status.update()

    def updateMaxTilt(self):
        new_max_tilt = dpg.get_value(self.menu_right.all_elements['max_tilt'])

        if abs(new_max_tilt - self.tgt_params.max_tilt) > 5:
            self.tgt_params.max_tilt = new_max_tilt

            # Reset target overlays
            for overlay in self.target_overlays:
                if dpg.does_item_exist(overlay):
                    dpg.delete_item(overlay)
            self.target_overlays = {}

            # Re-plot targets
            self.showTargets()

    def getTargetClassScores(self):
        """Gets class scores for cats in target_list."""

        for cat in self.tgt_params.target_list:
            for area in self.targets.areas:
                area.getClassScores(cat, self.segmentation.getMask([cat], unbinned=True))

    def changeArea(self, sender, app_data, user_data):
        """Changes target area of single target."""

        new_area = int(dpg.get_value(sender)) - 1
        self.targets.movePointToArea(user_data, new_area_id=new_area)

        dpg.hide_item("win_tgt")
        self.showTargets()
        self.showTargetAreaButtons()
        # Enable save button
        dpg.show_item(self.menu_right.all_elements["butsave"])
        
    def makeTrackDialogue(self, sender, app_data, user_data):
        """Creates dialogue for make tracking target."""

        dpg.hide_item("win_tgt")
        dpg.split_frame()
        gui.showInfoBox("Make tracking target", "Do you want to make this target the tracking target of the current area or create a new acquisition area?", callback=self.makeTrack, options=["Current area", "New area", "Cancel"], options_data=[[user_data, "old"], [user_data, "new"], False])

    def makeTrack(self, sender, app_data, user_data):
        """Makes selected target a tracking target of current or new area."""

        # Check for info box input
        if user_data and dpg.does_item_exist(user_data[0]):
            dpg.delete_item(user_data[0])
            dpg.split_frame()
        
        if not user_data[1]:
            return
        
        area_id, point_id = user_data[1][0]
        new_area = True if user_data[1][1] == "new" else False

        if new_area:
            # Get coords
            coords = self.targets.areas[area_id].points[point_id]

            # Remove point from old area
            self.targets.areas[area_id].removePoint(point_id)
            
            # Create new area with point
            self.targets.addTarget(coords, new_area=True)
            log("NOTE: Created new acquisition area!")
        else:
            # Move point to beginning of area
            self.targets.areas[area_id].makeTrack(point_id)
            log("NOTE: Made new tracking target!")

        self.showTargets()
        self.showTargetAreaButtons()
        # Enable save button
        dpg.show_item(self.menu_right.all_elements["butsave"])

    def optimizeTarget(self, sender, app_data, user_data):
        """Translates target according to loaded target mask."""
        
        # Check if segmentation is loaded
        if not self.overlay:
            log("ERROR: Please load class overlay first to optimize positioning of target.")
            return

        if user_data:
            area_id, point_id = user_data
            self.targets.areas[area_id].optimizeLocally(point_id, self.segmentation.getMask(self.overlay, unbinned=True))

            dpg.hide_item("win_tgt")
            self.showTargets()
            # Enable save button
            dpg.show_item(self.menu_right.all_elements["butsave"])

    def removeTarget(self, sender, app_data, user_data):
        """Deletes a single target."""

        if user_data:
            area_id, point_id = user_data
            self.targets.areas[area_id].removePoint(point_id)
            if len(self.targets.areas[area_id]) == 0:
                self.targets.areas.pop(area_id)
            dpg.hide_item("win_tgt")

            self.showTargets()
            self.showTargetAreaButtons()
            # Enable save button
            dpg.show_item(self.menu_right.all_elements["butsave"])

    def mergeAreas(self):
        """Merge all target areas."""

        self.targets.mergeAreas()

        self.showTargets()
        self.showTargetAreaButtons()
        # Enable save button
        dpg.show_item(self.menu_right.all_elements["butsave"])

    def splitAreas(self):
        """Splits targets into target areas using k-means clustering."""

        self.targets.splitArea(area_num=2)

        self.showTargets()
        self.showTargetAreaButtons()
        # Enable save button
        dpg.show_item(self.menu_right.all_elements["butsave"])

    def splitTargets(self):
        """Splits all targets among current tracking targets."""

        self.targets.splitAreaManual()

        self.showTargets()
        self.showTargetAreaButtons()
        # Enable save button
        dpg.show_item(self.menu_right.all_elements["butsave"])        

    def clearTargets(self, *args, plot_only=False):
        """Deletes all targets (or just clear from plot.)"""

        # Clear targets from plot
        self.plot.clearOverlays(delete_textures=not plot_only)
        self.plot.clearDragPoints()
        self.plot.clearSeries(labels=self.plot.getSeriesByKeyword("geo") + self.plot.getSeriesByKeyword("tgt") + [""])
        if not plot_only:
            # Release target overlay textures
            self.target_overlays = {}
            # Reinitialize targets object
            self.targets = Targets(map_dir=self.cur_dir, map_name=self.map_name, map_dims=self.loaded_map.img.shape, tgt_params=self.tgt_params, map_pix_size=self.loaded_map.pix_size)
            log("NOTE: Deleted targets!")

            # Enable save button
            dpg.show_item(self.menu_right.all_elements["butsave"])

    def clearGeoPoints(self):
        """Deletes all geo points."""

        # Clear all items with "geo" labels (needs empty list entry, otherwise all items will be cleared in case of no hits)
        self.plot.clearOverlays(labels=self.plot.getOverlaysByKeyword("geo") + [""], delete_textures=False)
        self.plot.clearDragPoints(labels=self.plot.getDragPointsByKeyword("geo") + [""])
        self.plot.clearSeries(labels=self.plot.getSeriesByKeyword("geo") + [""])

        self.targets.resetGeo()
        log("NOTE: Deleted geo points!")

        # Enable save button
        dpg.show_item(self.menu_right.all_elements["butsave"])

    def saveTargets(self):
        """Exports targets as json file."""

        # Get acquisition settings
        settings = self.getAcquisitionSettings()

        self.targets.exportTargets(settings)
        dpg.hide_item(self.menu_right.all_elements["butsave"])
        log("NOTE: Saved targets!")

    def markInspected(self, sender, app_data, user_data):
        """Creates inspected.txt file and locks down editing."""

        # Check for geo_points
        if self.targets and len(self.targets.areas[0].points) and not len(self.targets.areas[0].geo_points) and not user_data:
            log(f"WARNING: No geo points were selected!")
            gui.showInfoBox("WARNING", "No geo points were selected to measure the sample geometry. If you go ahead, the manual input for pretilt and rotation will be used!", callback=self.markInspected, options=["Continue", "Cancel"], options_data=[True, False])
            return
        # Close geo_points confirmation info box
        if user_data and dpg.does_item_exist(user_data[0]):
            dpg.delete_item(user_data[0])
            dpg.split_frame()
            # Cancel
            if not user_data[1]:
                return

        # Save targets if unsaved changes
        if dpg.is_item_shown(self.menu_right.all_elements["butsave"]):
            self.saveTargets()

        # Create empty inspected file
        (self.cur_dir / (self.map_name + "_inspected.txt")).touch()

        self.inspected = True

        # Update GUI
        self.menu_right.hide()
        dpg.show_item("inspected")

        # Update map window to reflected inspected status
        self.map_window.update(self.map_name, self.map_list, self.map_list_tgtnum)

        log("NOTE: Targets were marked as inspected!")

        # If there are not inspected maps left, load next map
        if not self.checkAllInspected():
            log("\n\nLoading next map...")
            self.selectMap(next_map=True)

    def markAllInspected(self, sender, app_data, user_data):
        """Marks all maps as inspected!"""

        for map_name in self.map_list:
            (self.cur_dir / (map_name + "_inspected.txt")).touch(exist_ok=True)

        self.inspected = True

        self.checkAllInspected()

    def checkAllInspected(self):
        """Checks if all lamella have been marked inspected and prompts GUI close."""

        # Check if all maps were inspected
        inspected = []
        for map_name in self.map_list:
            inspected.append((self.cur_dir / (map_name + "_inspected.txt")).exists())
        if all(inspected):
            log("NOTE: All lamella maps were inspected!")
            if self.auto_close:
                # Confirm closing
                gui.showInfoBox("FINISHED?", "All available lamellae were inspected. Are lamella maps still being acquired? If not, you can close this GUI.", callback=self.closeAllInspected, options=["Wait", "Close"], options_data=[False, True])
                return True
        else:
            return False

    def closeAllInspected(self, sender, app_data, user_data):
        """Closes GUI if all lamellae were inspected and user confirmed."""

        # Close GUI
        if user_data and user_data[1]:
            dpg.stop_dearpygui()
        else:
            # Close confirmation info box
            if dpg.does_item_exist(user_data[0]):
                dpg.delete_item(user_data[0])
                dpg.split_frame()        

    def getAcquisitionSettings(self):
        """Gets acquisition settings from input fields."""

        settings = {}
        # Only save settings if override checkbox was checked
        if dpg.get_value(self.menu_right.all_elements["acq_save"]):
            settings["startTilt"] = dpg.get_value(self.menu_right.all_elements["acq_start_tilt"])
            settings["minTilt"] = dpg.get_value(self.menu_right.all_elements["acq_min_tilt"])
            settings["maxTilt"] = dpg.get_value(self.menu_right.all_elements["acq_max_tilt"])
            settings["step"] = dpg.get_value(self.menu_right.all_elements["acq_step_tilt"])
            settings["pretilt"] = dpg.get_value(self.menu_right.all_elements["acq_pretilt"])
            settings["rotation"] = dpg.get_value(self.menu_right.all_elements["acq_rotation"])

        return settings
    
    def setAcquisitionSettings(self):
        """Sets acquisition settings in case they were loaded from file."""

        if self.targets.settings:
            dpg.set_value(self.menu_right.all_elements["acq_start_tilt"], self.targets.settings["startTilt"])
            dpg.set_value(self.menu_right.all_elements["acq_min_tilt"], self.targets.settings["minTilt"])
            dpg.set_value(self.menu_right.all_elements["acq_max_tilt"], self.targets.settings["maxTilt"])
            dpg.set_value(self.menu_right.all_elements["acq_step_tilt"], self.targets.settings["step"])
            dpg.set_value(self.menu_right.all_elements["acq_pretilt"], self.targets.settings["pretilt"])
            dpg.set_value(self.menu_right.all_elements["acq_rotation"], self.targets.settings["rotation"])

    def checkRunConditions(self):
        """Checks if mic and tgt params could be loaded."""

        if self.mic_params is None or self.model is None:
            if not self.checked_run_conditions:
                log("ERROR: No microscope parameter file found. Please launch the GUI in a SPACE_maps folder!")
                # Change exit callback to avoid askForSave method
                dpg.set_exit_callback(dpg.stop_dearpygui)
                gui.showInfoBox("ERROR", "No microscope parameter file found.\nPlease launch the GUI in a SPACE_maps folder!", callback=dpg.stop_dearpygui)
            self.checked_run_conditions = True
            return

        # Create popup if no maps have been found yet
        if not self.map_list:
            if not self.checked_map_files:
                gui.showInfoBox("NOTE", "No lamella map segmentations are available yet.\n\nReload after a few minutes to check for lamella map segmentations.", callback=self.manualMapListReload, options=["Reload", "Close"], options_data=[True, False])
            self.checked_map_files = True
            return

        # Reset mouse_move_handler
        if dpg.does_item_exist("check_run_condition"):
            dpg.delete_item("check_run_condition")

        # Set trigger help mouse_move_handler
        with dpg.handler_registry(tag="trigger_help"):
            dpg.add_mouse_move_handler(callback=self.triggerShowHelp)

    def manualMapListReload(self, sender, app_data, user_data):
        """Forces map list reload when prompted from info box."""

        # Close info box
        if dpg.does_item_exist(user_data[0]):
            dpg.delete_item(user_data[0])
            dpg.split_frame()

        if user_data[1]:
            self.updateMapList()
            # Reopen prompt if still no maps found
            if not self.map_list:
                self.checked_map_files = False
        else:
            dpg.stop_dearpygui()

    def makeTargetOverlay(self):
        """Generates target overlay textures."""

        # Return early if target overlays already exist
        if self.target_overlays:
            return

        # Get camera dims
        rec_dims = np.array(self.tgt_params.weight.shape)

        # TGT
        # Create canvas with size of stretched beam diameter
        tgt_overlay = np.zeros([int(self.model.beam_diameter), int(self.model.beam_diameter / np.cos(np.radians(self.tgt_params.max_tilt)))])
        canvas = Image.fromarray(tgt_overlay).convert('RGB')
        draw = ImageDraw.Draw(canvas)

        # Draw beam
        draw.ellipse((0, 0, tgt_overlay.shape[1] - 1, tgt_overlay.shape[0] - 1), outline="#ffd700", width=20)

        # Rotate tilt axis
        canvas = canvas.rotate(-self.mic_params.view_ta_rotation, expand=True)

        # Draw camera outline
        rect = ((canvas.width - rec_dims[1]) // 2, (canvas.height - rec_dims[0]) // 2, (canvas.width + rec_dims[1]) // 2, (canvas.height + rec_dims[0]) // 2)
        draw = ImageDraw.Draw(canvas)
        draw.rectangle(rect, outline="#578abf", width=20)

        # Convert to array
        tgt_overlay = np.array(canvas).astype(float) / 255

        # Draw camera outline for tracking target
        draw.rectangle(rect, outline="#c92b27", width=20)

        # Convert to array
        trk_overlay = np.array(canvas).astype(float) / 255

        # GEO
        # Create canvas for geo with non-stretched beam diameter
        geo_overlay = np.zeros([int(self.model.beam_diameter), int(self.model.beam_diameter)])
        canvas = Image.fromarray(geo_overlay).convert('RGB')
        draw = ImageDraw.Draw(canvas)

        # Draw beam and camera dims
        draw.ellipse((0, 0, geo_overlay.shape[1] - 1, geo_overlay.shape[0] - 1), outline="#ee8844", width=20)
        rect = ((canvas.width - rec_dims[1]) // 2, (canvas.height - rec_dims[0]) // 2, (canvas.width + rec_dims[1]) // 2, (canvas.height + rec_dims[0]) // 2)
        draw.rectangle(rect, outline="#ee8844", width=20)

        # Convert to array
        geo_overlay = np.array(canvas).astype(float) / 255

        # Make textures
        self.target_overlays["tgtdims"] = np.array(tgt_overlay.shape)[:2]
        alpha = np.zeros(tgt_overlay.shape[:2])
        alpha[np.sum(tgt_overlay, axis=-1) > 0] = 1
        tgt_overlay_image = np.ravel(np.dstack([tgt_overlay, alpha]))
        trk_overlay_image = np.ravel(np.dstack([trk_overlay, alpha]))

        self.target_overlays["geodims"] = np.array(geo_overlay.shape)[:2]
        alpha = np.zeros(geo_overlay.shape[:2])
        alpha[np.sum(geo_overlay, axis=-1) > 0] = 1
        geo_overlay_image = np.ravel(np.dstack([geo_overlay, alpha]))

        with dpg.texture_registry():
            self.target_overlays["target"] = dpg.add_static_texture(width=int(self.target_overlays["tgtdims"][1]), height=int(self.target_overlays["tgtdims"][0]), default_value=tgt_overlay_image)
            self.target_overlays["track"] = dpg.add_static_texture(width=int(self.target_overlays["tgtdims"][1]), height=int(self.target_overlays["tgtdims"][0]), default_value=trk_overlay_image)
            self.target_overlays["geo"] = dpg.add_static_texture(width=int(self.target_overlays["geodims"][1]), height=int(self.target_overlays["geodims"][0]), default_value=geo_overlay_image)

    def askForSave(self):
        """Opens popup if there are unsaved changes."""

        # Check for unsaved changes
        if dpg.is_item_shown(self.menu_right.all_elements["butsave"]):
            gui.showInfoBox("WARNING", "There are unsaved changes to your targets!", callback=self.saveAndClose, options=["Save", "Discard"], options_data=[True, False])
        else:

            # Check inspected status and give warning when some but not all targets have been inspected
            inspected = []
            for map_name in self.map_list:
                inspected.append((self.cur_dir / (map_name + "_inspected.txt")).exists())
            if any(inspected) and not all(inspected):
                gui.showInfoBox("WARNING", "Not all targets have been marked as inspected. Are you sure you want to exit?", callback=self.confirmClose, options=["Exit", "Cancel"], options_data=[True, False])
            else:
                dpg.stop_dearpygui()

    def saveAndClose(self, sender, app_data, user_data):
        """Saves targets and closes GUI."""

        # Save targets if user clicked Save
        if user_data[1]:
            self.saveTargets()

        # Change exit callback to avoid infinite loop
        dpg.set_exit_callback(dpg.stop_dearpygui)
        dpg.stop_dearpygui()

    def confirmClose(self, sender, app_data, user_data):
        """Closes GUI or keeps it open."""

        if user_data[1]:
            # Change exit callback to avoid infinite loop
            dpg.set_exit_callback(dpg.stop_dearpygui)
            dpg.stop_dearpygui()
        else:
            # Close info box
            if dpg.does_item_exist(user_data[0]):
                dpg.delete_item(user_data[0])
                dpg.split_frame()

    def toggleAdvanced(self):
        """Shows advanced menu options."""

        self.menu_left.toggleAdvanced()
        self.menu_right.toggleAdvanced()

    def triggerShowHelp(self):
        """Triggers show help once."""

        if dpg.does_item_exist("trigger_help"):
            dpg.delete_item("trigger_help")

        # Show help
        self.showHelp()

    @staticmethod
    def showHelp():
        """Shows popup with shortcuts."""

        message = "Load a lamella map from the list to\ninspect or select targets and geo points.\n\n[Confirm inspection] when you are done!\n\n\n"
        message += "Controls:\n\n"
        message += "Shift + left click      Add new target\n"
        message += "Right click on target   Open menu\n"
        message += "Middle click            Add geo point\n"
        message += "\n\n"
        message += "Keyboard shortcuts:\n\n"
        message += "A     Toggle advanced settings\n"
        message += "H     Show help\n"
        message += "N     Load next map\n"
        message += "R     CLEM registration\n"
        message += "Space Show available maps\n"

        gui.showInfoBox("Help", message)

#################################################################################


    def __init__(self, path="", auto_close=False) -> None:
        log(f"\n########################################\nRunning SPACEtomo Target Selection GUI\n########################################\n")

        # Automatically close GUI after map was inspected
        self.auto_close = auto_close

        # Make logo
        self.logo_dims = gui.makeLogo()

        self.configureHandlers()
        self.configureThemes()

        # Create plot object to hold all entities
        self.plot = Plot("plot")

        # Initialise globals
        self.menu_left = None
        self.menu_right = None
        self.status = None
        
        # Check given path
        if path:
            path = Path(path)
            if not path.is_dir():
                path = path.parent
            self.cur_dir = Path(path)
        else:
            self.cur_dir = Path.cwd()

        # Maps and targets
        self.map_list = []                          # List of segmentation files found in folder
        self.map_name = ""                          # Name of currently loaded map
        self.loaded_map = None                      # MMap object
        self.overlay = []                           # Classes currently shown in overlay
        self.targets = None                         # Targets object containing points, scores and geo_points for all areas 
        self.target_overlays = {}                   # Textures for target overlays
        self.map_list_tgtnum = []                   # List of target numbers for each map
        self.inspected = False                      # Inspection state of current map

        # Parameters
        self.mic_params = None                      # Mic parameter object
        self.model = None                           # Segmentation model object
        self.tgt_params = None                      # Target parameter object
        self.loadParameterFiles()

        # One-time calls
        self.checked_run_conditions = False           # Needed to show error popup only once
        self.checked_point_files = False            # Needed to only prompt user once to load found point files
        self.checked_map_files = False              # Needed to only prompt user once to reload maps

        # Settings
        self.use_thumbnails = False

        # Developer settings
        self.thumbnail_size = (100, 100)            # Dims of thumbnails for map window

    def loadParameterFiles(self):
        """Loads parameters saved by SPACEtomo."""

        # Find all maps with segmentation
        thumbnail_size = (100, 100)
        self.updateMapList(list_only=True)

        # Check if mic_params exist
        if self.cur_dir.joinpath("mic_params.json").exists():
            log("NOTE: Microscope parameters found.")
            # Instantiate mic params from settings
            self.mic_params = space_ext.MicParams_ext(self.cur_dir)        
            # Load model
            self.model = space_ext.MMModel()
            self.model.setDimensions(self.mic_params)

            # Check if tgt_params exist
            if self.cur_dir.joinpath("tgt_params.json").exists():
                log("NOTE: Target parameters found.")
                # Instantiate tgt params from settings
                self.tgt_params = space_ext.TgtParams(file_dir=self.cur_dir, MM_model=self.model)
            else:
                # Use defaults
                self.tgt_params = space_ext.TgtParams(["lamella"], ["black", "white", "crack", "ice"], self.model, False, False, 0.3, 0.5, 60, self.mic_params, False, 10)

    def configureHandlers(self):
        """Sets up dearpygui registries and handlers."""

        # Create file dialogues

        # Create event handlers
        with dpg.handler_registry(tag="mouse_handler"):
            dpg.add_mouse_click_handler(callback=self.mouseClick)
            dpg.add_mouse_release_handler(button=dpg.mvMouseButton_Left, callback=self.mouseRelease)
            #m_release_left = dpg.add_mouse_release_handler(button=dpg.mvMouseButton_Left, callback=tgt_tgtUpdate)
            dpg.add_mouse_move_handler(tag="check_run_condition", callback=self.checkRunConditions)

            # Shortcuts
            dpg.add_key_press_handler(dpg.mvKey_H, callback=self.showHelp)
            dpg.add_key_press_handler(dpg.mvKey_N, callback=lambda: self.selectMap(next_map=True))
            dpg.add_key_press_handler(dpg.mvKey_A, callback=self.toggleAdvanced)
            dpg.add_key_press_handler(dpg.mvKey_Spacebar, callback=lambda: self.map_window.show())
            dpg.add_key_press_handler(dpg.mvKey_R, callback=lambda: self.lm_window.show() if not dpg.is_item_shown(self.lm_window.window) else self.lm_window.hide())

        with dpg.item_handler_registry(tag="point_tooltip_handler"):
            dpg.add_item_hover_handler(callback=self.targetTooltip)

        with dpg.item_handler_registry(tag="class_input_handler"):
            dpg.add_item_clicked_handler(button=dpg.mvMouseButton_Left, callback=self.openClassSelection)

        dpg.set_viewport_resize_callback(callback=lambda: gui.window_size_change(self.logo_dims))

    @staticmethod
    def configureThemes():
        """Sets up dearpygui themes."""

        gui.configureGlobalTheme()

        # Color themes for target areas plots
        for c, color in enumerate(gui.THEME_COLORS):
            with dpg.theme(tag="scatter_theme" + str(c)):
                with dpg.theme_component(dpg.mvScatterSeries):
                    dpg.add_theme_color(dpg.mvPlotCol_Line, color, category=dpg.mvThemeCat_Plots)

        # Theme for temporary drag point
        with dpg.theme(tag="drag_scatter_theme"):
            with dpg.theme_component(dpg.mvScatterSeries):
                dpg.add_theme_color(dpg.mvPlotCol_Line, (200, 0, 0, 255), category=dpg.mvThemeCat_Plots)
                dpg.add_theme_style(dpg.mvPlotStyleVar_Marker, dpg.mvPlotMarker_Circle, category=dpg.mvThemeCat_Plots)
                dpg.add_theme_style(dpg.mvPlotStyleVar_MarkerSize, 7, category=dpg.mvThemeCat_Plots)

        # Theme for tilt axis plot
        with dpg.theme(tag="axis_theme"):
            with dpg.theme_component(dpg.mvLineSeries):
                dpg.add_theme_color(dpg.mvPlotCol_Line, (255, 255, 255, 64), category=dpg.mvThemeCat_Plots)

    def show(self):
        """Structures and launches main window of GUI."""

        # Setup window
        dpg.create_viewport(title="SPACEtomo Target Selection", disable_close=True)
        dpg.setup_dearpygui()

        # Create main window
        with dpg.window(label="GUI", tag="GUI", no_scrollbar=True, no_scroll_with_mouse=True):

            with dpg.table(header_row=False):
                dpg.add_table_column(init_width_or_weight=200, width_fixed=True)
                dpg.add_table_column()
                dpg.add_table_column(init_width_or_weight=200, width_fixed=True)

                with dpg.table_row():
                    with dpg.table_cell(tag="tblleft"):
                        dpg.add_text(default_value="Load your MM map", tag="l1", color=gui.COLORS["heading"])

                        self.menu_left = Menu()

                        self.menu_left.newRow(tag="mapcombo", separator=False, locked=False)
                        self.menu_left.addCombo(tag="map", label=f"({len(self.map_list)} maps)", combo_list=self.map_list)
                        self.updateMapList()
                        
                        self.menu_left.newRow(tag="load", horizontal=True, separator=False, locked=False)
                        self.menu_left.addButton(tag="butload", label="Load map", callback=self.selectMap)
                        self.menu_left.addButton(tag="butnext", label="Next", callback=lambda: self.selectMap(next_map=True), show=bool(self.map_list))
                        self.menu_left.addButton(tag="butgrid", label="[]", callback=lambda: self.map_window.show())

                        self.menu_left.newRow(tag="advload", separator=False, locked=False, advanced=True)
                        self.menu_left.addInput(tag="inpbin", label="Binning", value=1)

                        self.menu_left.newRow(tag="windows", horizontal=True, separator=False, locked=False)
                        self.menu_left.addButton(tag="butlamgui", label="Grid map", callback=lambda: utils.guiProcess("lamella"))
                        self.menu_left.addButton(tag="butflm", label="CLEM", callback=lambda: self.lm_window.show())

                        self.menu_left.newRow(tag="filters", separator=False, locked=True, advanced=False)
                        self.menu_left.addButton(tag="butfilter", label="Filter map", callback=self.filterMap)

                        self.menu_left.newRow(tag="reacquire", separator=False, locked=True, advanced=True)
                        self.menu_left.addButton(tag="butreacq", label="Reacquire", callback=self.reacquireMap)

                        self.menu_left.newRow(tag="class_list", separator=False, locked=True)
                        self.menu_left.addText(tag="class_heading", value="\nClasses", color=gui.COLORS["heading"])
                        if self.model is not None:
                            for key in self.model.categories.keys():
                                self.menu_left.addCheckbox(tag=key, label=key, value=False)

                        self.menu_left.newRow(tag="class_buttons", horizontal=True, separator=False, locked=True)
                        self.menu_left.addButton(tag="clsmask", label="Create overlay", callback=self.loadOverlay)
                        self.menu_left.addButton(tag="clsapply", label="Apply", callback=self.applyClassSelection, user_data=[self.menu_left, None])

                        self.status = gui.StatusLine()

                    with dpg.table_cell(tag="tblplot"):
                        self.plot.makePlot(x_axis_label="x [µm]", y_axis_label="y [µm]", width=-1, height=-1, equal_aspects=True, no_menus=True, crosshairs=True, pan_button=dpg.mvMouseButton_Right, no_box_select=True)

                    with dpg.table_cell(tag="tblright"):
                        dpg.add_text(default_value="Target selection", tag="r1", color=gui.COLORS["heading"])

                        if self.mic_params is not None:
                            self.menu_right = Menu()

                            self.menu_right.newRow(tag="settings1", separator=False)
                            self.menu_right.addText(tag="target_list_label", value="Target classes:")
                            self.menu_right.addInput(tag="target_list", label="", value=",".join(self.tgt_params.target_list), width=-1)
                            dpg.bind_item_handler_registry(self.menu_right.all_elements["target_list"], "class_input_handler")

                            self.menu_right.newRow(tag="settings2", separator=False)
                            self.menu_right.addText(tag="avoid_list_label", value="Avoid classes:")
                            self.menu_right.addInput(tag="avoid_list", label="", value=",".join(self.tgt_params.penalty_list), width=-1)
                            dpg.bind_item_handler_registry(self.menu_right.all_elements["avoid_list"], "class_input_handler")

                            self.menu_right.newRow(tag="settings3", separator=False)
                            self.menu_right.addText(tag="settings_heading", value="\nTargeting options:")
                            #self.menu_right.addInput(tag="target_score_threshold", label="Score threshold", value=self.tgt_params.threshold)
                            self.menu_right.addSlider(tag="target_score_threshold", label="Score threshold", value=float(self.tgt_params.threshold), value_range=[0, 1], width=75)
                            #self.menu_right.addInput(tag="penalty_weight", label="Penalty weight", value=self.tgt_params.penalty)
                            self.menu_right.addSlider(tag="penalty_weight", label="Penalty weight", value=float(self.tgt_params.penalty), value_range=[0, 1], width=75)
                            #self.menu_right.addInput(tag="max_tilt", label="Max. tilt angle", value=self.tgt_params.max_tilt, callback=self.updateMaxTilt)
                            self.menu_right.addSlider(tag="max_tilt", label="Max. tilt angle", value=int(self.tgt_params.max_tilt), value_range=[0, 80], width=75, callback=self.updateMaxTilt)
                            #self.menu_right.addInput(tag="IS_limit", label="Image shift limit", value=self.mic_params.IS_limit)
                            self.menu_right.addSlider(tag="IS_limit", label="Image shift limit", value=int(self.mic_params.IS_limit), value_range=[5, 20], width=75)
                            self.menu_right.addCheckbox(tag="sparse_targets", label="Sparse targets", value=self.tgt_params.sparse)
                            self.menu_right.addCheckbox(tag="target_edge", label="Target edge", value=self.tgt_params.edge)
                            self.menu_right.addCheckbox(tag="extra_tracking", label="Extra tracking", value=self.tgt_params.sparse)

                            self.menu_right.newRow(tag="settings4", separator=False)
                            self.menu_right.addButton(tag="butselect", label="Auto select targets", callback=self.runTargetSelection)
                            self.menu_right.addText(tag="rsp1", value="")

                            self.menu_right.newRow(tag="areas", separator=False)
                            self.menu_right.addButton(tag="butsplit", label="Split target areas", callback=self.splitAreas)
                            self.menu_right.addButton(tag="butdist", label="Redistribute targets", callback=self.splitTargets)
                            self.menu_right.addButton(tag="butmerge", label="Merge target areas", callback=self.mergeAreas)

                            self.menu_right.newRow(tag="delete", separator=False)
                            self.menu_right.addButton(tag="butdelete", label="Delete targets", callback=self.clearTargets)
                            self.menu_right.addButton(tag="butdeletegeo", label="Delete geo points", callback=self.clearGeoPoints)

                            self.menu_right.newRow(tag="acquisition1", separator=False)
                            self.menu_right.addText(tag="acq_heading", value="\nAcquisition settings", color=gui.COLORS["heading"])
                            self.menu_right.addText(tag="tilt_label1", value="Tilt angles [deg]:")

                            self.menu_right.newRow(tag="acquisition2", horizontal=True, separator=False)
                            self.menu_right.addInput(tag="acq_start_tilt", label="Start  ", value=0, width=34, callback=lambda: dpg.set_value(self.menu_right.all_elements["acq_save"], True))
                            self.menu_right.addInput(tag="acq_step_tilt", label="Step   ", value=3, width=34, callback=lambda: dpg.set_value(self.menu_right.all_elements["acq_save"], True))
                            
                            self.menu_right.newRow(tag="acquisition3", horizontal=True, separator=False)
                            self.menu_right.addInput(tag="acq_min_tilt", label="Min    ", value=-60, width=34, callback=lambda: dpg.set_value(self.menu_right.all_elements["acq_save"], True))
                            self.menu_right.addInput(tag="acq_max_tilt", label="Max    ", value=60, width=34, callback=lambda: dpg.set_value(self.menu_right.all_elements["acq_save"], True))

                            self.menu_right.newRow(tag="acquisition4", separator=False)
                            self.menu_right.addText(tag="geo_label1", value="Fallback geometry [deg]:")

                            self.menu_right.newRow(tag="acquisition5", horizontal=True, separator=False)
                            self.menu_right.addInput(tag="acq_pretilt", label="Pretilt", value=0, width=34, callback=lambda: dpg.set_value(self.menu_right.all_elements["acq_save"], True))
                            self.menu_right.addInput(tag="acq_rotation", label="Rotation", value=0, width=34, callback=lambda: dpg.set_value(self.menu_right.all_elements["acq_save"], True))

                            self.menu_right.newRow(tag="acquisition6", horizontal=True, separator=False)
                            self.menu_right.addCheckbox(tag="acq_save", label="Override script settings")

                            self.menu_right.newRow(tag="buttons1", separator=False)
                            self.menu_right.addText(tag="save_heading", value="\n\nSave targets", color=gui.COLORS["heading"])

                            self.menu_right.newRow(tag="buttons2", separator=False)
                            self.menu_right.addButton(tag="butsave", label="Save", callback=self.saveTargets, show=False)
                            self.menu_right.addButton(tag="butins", label="Confirm inspection", callback=self.markInspected, theme="large_btn_theme")
                            self.menu_right.addButton(tag="butins", label="Confirm all maps", callback=self.markAllInspected)

                            self.menu_right.hide()

                            dpg.add_text(tag="inspected", default_value="\nTargets for this map were\nalready marked as inspected.\n(Editing disabled)", color=gui.COLORS["heading"], show=False)
                        else:
                            dpg.add_text(tag="rerr", default_value="No microscope parameters\nfile found!", color=gui.COLORS["error"])


            # Create tooltips
            with dpg.tooltip("l1", delay=0.5):
                dpg.add_text("Select and load a map generated by SPACEtomo.")
            with dpg.tooltip(self.menu_left.all_elements["class_heading"], delay=0.5):
                dpg.add_text("Choose a class to be displayed as overlay.")
            with dpg.tooltip(self.menu_left.all_elements["butreacq"], delay=0.5):
                dpg.add_text("This will require you to restart the SPACEtomo run!")
            with dpg.tooltip(self.menu_left.all_elements["clsmask"], delay=0.5):
                dpg.add_text("Create overlay of selected classes. (This can take a few seconds.)")
            with dpg.tooltip(self.menu_left.all_elements["clsapply"], delay=0.5):
                dpg.add_text("Apply selected classes to target list.")

            with dpg.tooltip("tblplot", delay=0.5, hide_on_activity=True):
                dpg.add_text(default_value="", color=gui.COLORS["heading"], tag="tt_heading")
                dpg.add_text(default_value="", tag="tt_text")
            dpg.bind_item_handler_registry("plot", "point_tooltip_handler")

            if self.mic_params is not None:
                with dpg.tooltip("r1", delay=0.5):
                    dpg.add_text("Select targets based on segmentation.")
                with dpg.tooltip(self.menu_right.all_elements["target_list"], delay=0.5):
                    dpg.add_text("List of target classes (comma separated). For exhaustive acquisition use \"lamella\".")
                with dpg.tooltip(self.menu_right.all_elements["avoid_list"], delay=0.5):
                    dpg.add_text("List of classes to avoid (comma separated).")
                with dpg.tooltip(self.menu_right.all_elements["target_score_threshold"], delay=0.5):
                    dpg.add_text("Score threshold [0-1] below targets will be excluded.")
                with dpg.tooltip(self.menu_right.all_elements["penalty_weight"], delay=0.5):
                    dpg.add_text("Relative weight of avoided classes to target classes.")
                with dpg.tooltip(self.menu_right.all_elements["max_tilt"], delay=0.5):
                    dpg.add_text("Maximum tilt angle [degrees] to consider electron beam exposure.")
                with dpg.tooltip(self.menu_right.all_elements["IS_limit"], delay=0.5):
                    dpg.add_text("Image shift limit [µm] for PACEtomo acquisition. If targets are further apart, target area will be split.")
                with dpg.tooltip(self.menu_right.all_elements["sparse_targets"], delay=0.5):
                    dpg.add_text("Target positions will be initialized only on target classes and refined independently (instead of grid based target target setup to minimize exposure overlap).")
                with dpg.tooltip(self.menu_right.all_elements["target_edge"], delay=0.5):
                    dpg.add_text("Targets will be centered on edge of segmented target instead of maximising coverage.")
                with dpg.tooltip(self.menu_right.all_elements["extra_tracking"], delay=0.5):
                    dpg.add_text("An extra target will be placed centrally for tracking.")

                with dpg.tooltip(self.menu_right.all_elements["butdelete"], delay=0.5):
                    dpg.add_text("Delete all targets and setup from scratch. (This cannot be undone!)")
                with dpg.tooltip(self.menu_right.all_elements["butins"], delay=0.5):
                    dpg.add_text("Mark targets as inspected. (No more changes can be made.)")

            # Show logo
            dpg.add_image("logo", pos=(10, dpg.get_viewport_client_height() - 40 - self.logo_dims[0]), tag="logo_img")
            dpg.add_text(default_value="SPACEtomo", pos=(10 + self.logo_dims[1] / 2 - (30), dpg.get_viewport_client_height() - 40 - self.logo_dims[0] / 2), tag="logo_text")
            dpg.add_text(default_value="v" + __version__, pos=(10 + self.logo_dims[1] / 2 - (30), dpg.get_viewport_client_height() - 27 - self.logo_dims[0] / 2), tag="version_text")

        # Make window for map thumbnails
        self.map_window = MapWindow(self.cur_dir, self.map_name, self.map_list, self.map_list_tgtnum, self.selectMap, self.use_thumbnails, self.thumbnail_size)
        self.map_window.makeMapTable()

        # LM window
        self.lm_window = FlmWindow(self.plot)

        # Make window for target editing menu
        with dpg.window(label="Target", tag="win_tgt", no_scrollbar=True, no_scroll_with_mouse=True, popup=True, show=False):
            self.menu_tgt = Menu()
            self.menu_tgt.newRow(tag="heading", separator=False, locked=False)
            self.menu_tgt.addText(tag="heading_txt", value="Target", color=gui.COLORS["heading"])
            self.menu_tgt.newRow(tag="info", separator=False)
            self.menu_tgt.addText(tag="info_txt")
            
            self.menu_tgt.newRow(tag="area", horizontal=True, separator=False)
            self.menu_tgt.addCombo(tag="area", label="Target area", callback=self.changeArea, width=30)

            self.menu_tgt.newRow(tag="buttons", separator=False)
            self.menu_tgt.addButton(tag="btn_del", label="Delete", callback=self.removeTarget)
            self.menu_tgt.addButton(tag="btn_trk", label="Make tracking target", callback=self.makeTrackDialogue)
            self.menu_tgt.addButton(tag="btn_opt", label="Optimize position", callback=self.optimizeTarget)

        # Make window for class list selection
        with dpg.window(label="Selection", tag="win_sel", no_scrollbar=True, no_scroll_with_mouse=True, popup=True, show=False):
            self.menu_sel = Menu()
            self.menu_sel.newRow(tag="heading", separator=False, locked=False)
            self.menu_sel.addText(tag="heading", value="Classes", color=gui.COLORS["heading"])
            if self.model is not None:
                for k, key in enumerate(self.model.categories.keys()):
                    if k % 3 == 0:
                        self.menu_sel.newRow(tag=str(k), horizontal=True, separator=False, locked=False)
                    self.menu_sel.addCheckbox(tag=key, label=key, value=False, callback=self.applyClassSelection)

        dpg.bind_theme("global_theme")

        dpg.set_exit_callback(self.askForSave)

        dpg.set_primary_window("GUI", True)
        dpg.show_viewport()

        # Render loop
        next_update = time.time() + 30
        while dpg.is_dearpygui_running():

            # Recheck folder for segmentation every minute
            now = time.time()
            if now > next_update:
                next_update = now + 30
                # Check if new maps were finished
                if self.updateMapList():
                    self.map_window.update(self.map_name, self.map_list, self.map_list_tgtnum)
                # Check if new target selection was finished
                self.checkPointFiles()

            dpg.render_dearpygui_frame()



#!/usr/bin/env python
# ===================================================================
# ScriptName:   buf
# Purpose:      Interface for SerialEM buffer.
#               More information at http://github.com/eisfabian/SPACEtomo
# Author:       Fabian Eisenstein
# Created:      2024/08/14
# Revision:     v1.2
# Last Change:  2024/09/04: added new polygon via nav to avoid listToSEMarray, fixed rollBuffers
#               2024/08/23: added BufferDummy
#               2024/08/19: added getCropImage from sem.py and addNavPoint
#               2024/08/16: changed nav_id convention for consistency, added check before file saving
#               2024/08/15: added buffer rolling, added scaling for addNavBoxes 
# ===================================================================

# Check SerialEM availability
try:
    import serialem as sem
    # Check if PySEMSocket is connected
    try:
        sem.Delay(0, "s")
        SERIALEM = True
    except sem.SEMmoduleError:
        SERIALEM = False
except ModuleNotFoundError:
    SERIALEM = False

import sys
import time
import json
import weakref
import numpy as np
import concurrent.futures
from pathlib import Path
from PIL import Image
Image.MAX_IMAGE_PIXELS = None           # removes limit for large images
from skimage import transform, exposure

from SPACEtomo import config
from SPACEtomo.modules import utils
from SPACEtomo.modules.utils import log

class Buffer:

    # Keep track of all instances to remove buffer letters when letters are overwritten
    instances = []

    # Get rolling buffers
    if SERIALEM:
        rolling_buffers = [chr(i + 65) for i in range(int(sem.ReportUserSetting("BuffersToRollOnAcquire")) + 1)]
    else:
        rolling_buffers = []

    # Reference to active nav
    nav = None

    def __init__(self, buffer="A", nav_id=None, img=None) -> None:
        """Buffer can be initialized in three ways:
            1. Using a navigator ID of a map.
            2. Using a buffer containing an image.
            3. Using a buffer and an image to place into it."""

        # Initialize attributes
        self.img = None             # Image as numpy array
        self.img_mod = None         # Modified image as numpy array

        # Load nav map
        if nav_id is not None:
            self.nav_id = nav_id
            sem.LoadOtherMap(self.nav_id + 1, buffer)
            self.buf, *_ = sem.ReportCurrentBuffer()
            self.buf_mod = self.buf

        # Use loaded buffer
        elif buffer and not img:
            self.nav_id = None
            self.buf = buffer
            self.buf_mod = self.buf
            sem.Show(self.buf)
        
        # Use provided image
        elif buffer and img:
            self.nav_id = None
            self.buf = buffer
            self.buf_mod = self.buf
            self.img = img
            self.imgToBuffer()
            # possibly sem.Show(self.buf) might be necessary to change to filled buffer

        else:
            raise ValueError("Need buffer letter (and optional image) or navigator item ID to initialize!")
        
        # Get image info
        img_prop = sem.ImageProperties(self.buf)
        self.shape = np.array([img_prop[0], img_prop[1]])
        self.binning = int(img_prop[2])
        self.pix_size = float(img_prop[4])

        # Get timestamp of image to make sure no other image was loaded in buffer when returning
        self.timestamp = sem.GetVariable("imageTickTime")

        # Add weak reference to instance list
        self.__class__.instances.append(weakref.ref(self))

    def show(self, reload=False):
        """Show buffer and assert that buffer is unchanged."""

        sem.Show(self.buf)

        # Get image info
        img_prop = sem.ImageProperties(self.buf)

        # Get timestamp
        timestamp = sem.GetVariable("imageTickTime")

        if timestamp != self.timestamp or reload:
            # Reload from map if buffer was overwritten
            if self.nav_id:
                sem.LoadOtherMap(self.nav_id + 1, self.buf)
            # Or place image in buffer
            elif self.img is not None:
                self.imgToBuffer()
            else:
                log(f"ERROR: Buffer is no longer available!")

    def getMetadata(self, key):
        """Gets specific mdoc entry."""

        return sem.ReportMetadataValues(self.buf, key)

    def imgFromBuffer(self, mod=False, reload=False):
        """Converts buffer to image as numpy array."""

        # Wrap in loop because it sometimes fails and add timeout call in case it freezes without error
        for i in range(10):
            try:
                if not mod:
                    if self.img is None or reload:
                        log(f"DEBUG: Getting image from buffer {self.buf}...")
                        self.img = utils.timeoutCall(lambda: np.asarray(sem.bufferImage(self.buf)))
                else:
                    if self.img_mod is None or reload:
                        log(f"DEBUG: Getting modified image from buffer {self.buf_mod}...")
                        self.img_mod = utils.timeoutCall(lambda: np.asarray(sem.bufferImage(self.buf_mod)))
                break
            except (sem.SEMmoduleError, SystemError):
                if i < 9:
                    log(f"WARNING: Could not obtain image from buffer. Trying again... [{i + 1}]")
                    time.sleep(1)
                else:
                    log(f"ERROR: Could not obtain image from buffer! There might be a problem with SerialEM.")
                    sys.exit()

    def imgToBuffer(self, mod=False, buffer=None):
        """Shows img in buffer."""

        if not buffer:
            buffer = self.buf

        if mod:
            sem.PutImageInBuffer(self.img_mod, buffer, *self.img_mod.shape, self.buf)
        else:
            sem.PutImageInBuffer(self.img, buffer, *self.img.shape, self.buf)

        # Roll buffers only if put in A (see script commands page for PutImageInBuffer)
        if buffer == "A":
            self.rollBuffers(skip=["A"])
        
    def saveImg_old(self, file_path, target_pix_size=None, save_meta_data=True, overwrite=False):
        """Saves buffer image as png image (optionally rescales it)."""

        if file_path.exists():
            if not overwrite:
                log(f"WARNING: {file_path.name} already exists! Skipping saving.")
                return
            else:
                log(f"WARNING: {file_path.name} already exists! It will be overwritten.")

        if target_pix_size:
            # Rescale image
            rescale_factor = target_pix_size / self.pix_size
            log(f"DEBUG: Rescaling image by factor: {rescale_factor}")
            if rescale_factor > 1:
                # Use reduce in case of downscaling (faster)
                self.reduceBuffer(rescale_factor)
                self.imgFromBuffer(mod=True, reload=True)
            else:
                # Use rescale in case of upscaling
                self.imgFromBuffer()
                self.rescaleImage(rescale_factor)
            # Use modified image
            img = self.img_mod
        else:
            # Use unmodified image
            self.imgFromBuffer()
            img = self.img

        # Check intensity range and rescale if not between 0 and 255
        log(f"DEBUG: Image min/max: {np.min(img)}/{np.max(img)}")
        if np.min(img) < 0 or not 200 < np.max(img) <= 255:
            log(f"DEBUG: Rescaling image intensity to out_range=(0, 255)") 
            img = exposure.rescale_intensity(img, out_range=(0, 255)).astype(np.uint8)

        log(f"DEBUG: Saving image to file...")
        Image.fromarray(img).save(file_path)

        if save_meta_data:
            log(f"DEBUG: Saving meta data to {file_path.parent / (file_path.stem + '.json')}")
            self.saveMetaData(file_path=file_path.parent / (file_path.stem + ".json"), target_pix_size=target_pix_size)

    def saveImg(self, file_path, target_pix_size=None, save_meta_data=True, overwrite=False):
        """Saves buffer image as png image (optionally rescales it)."""

        if file_path.exists():
            if not overwrite:
                log(f"WARNING: {file_path.name} already exists! Skipping saving.")
                return
            else:
                log(f"WARNING: {file_path.name} already exists! It will be overwritten.")

        if target_pix_size:
            # Rescale image
            rescale_factor = target_pix_size / self.pix_size
            log(f"DEBUG: Rescaling image by factor: {rescale_factor}")
            if rescale_factor > 1:
                # Use reduce in case of downscaling (faster)
                self.reduceBuffer(rescale_factor)
                self.imgFromBuffer(mod=True, reload=True)
            else:
                # Use rescale in case of upscaling
                self.imgFromBuffer()
                self.rescaleImage(rescale_factor)
            # Use modified image
            img = self.img_mod
        else:
            # Use unmodified image
            self.imgFromBuffer()
            img = self.img

        # Internal function that performs the actual save operation
        def save_operation(img):
            # Check intensity range and rescale if not between 0 and 255
            log(f"DEBUG: Image min/max: {np.min(img)}/{np.max(img)}")
            if np.min(img) < 0 or not 200 < np.max(img) <= 255:
                log(f"DEBUG: Rescaling image intensity to out_range=(0, 255)") 
                img = exposure.rescale_intensity(img, out_range=(0, 255)).astype(np.uint8)

            log(f"DEBUG: Saving image to file...")
            Image.fromarray(img).save(file_path)
            log(f"DEBUG: Finished saving {file_path}")

        # Start the background thread to run save_operation
        executor = concurrent.futures.ThreadPoolExecutor()
        future = executor.submit(save_operation, img)

        if save_meta_data:
            log(f"DEBUG: Saving meta data to {file_path.parent / (file_path.stem + '.json')}")
            self.saveMetaData(file_path=file_path.parent / (file_path.stem + ".json"), target_pix_size=target_pix_size)

        return future

    def saveMetaData(self, file_path, target_pix_size=None):
        """Saves meta data for buffer image as json file."""

        meta_data = {"datetime": time.strftime("%Y-%m-%d %H:%M:%S", time.localtime()), "img_size": self.shape}

        if target_pix_size:
            meta_data["pix_size"] = target_pix_size
            meta_data["img_size"] = np.array(self.img_mod.shape)
            scaling_factor = target_pix_size / self.pix_size
        else:
            meta_data["pix_size"] = self.pix_size
            scaling_factor = 1

        # Get additional data from nav
        if self.nav and self.nav_id:
            meta_data["stage_coords"] = np.array(self.nav.items[self.nav_id].entries["StageXYZ"])
            if "MapScaleMat" in self.nav.items[self.nav_id].entries.keys():
                matrix = np.array([float(val) for val in self.nav.items[self.nav_id].entries["MapScaleMat"]]).reshape((2, 2))
                meta_data["s2img_matrix"] = matrix / scaling_factor
            if "MapFile" in self.nav.items[self.nav_id].entries.keys():
                meta_data["original_map"] = Path(" ".join(self.nav.items[self.nav_id].entries["MapFile"])).as_posix()
            if "MapTiltAngle" in self.nav.items[self.nav_id].entries.keys():
                meta_data["tilt_angle"] = float(self.nav.items[self.nav_id].entries["MapTiltAngle"][0])
            # Can be extended if necessary

        with open(file_path, "w") as f:
            json.dump(meta_data, f, indent=4, default=utils.convertArray)

    def rescaleImage(self, factor):
        """Rescales image using skimage (slower, but can also upscale)."""

        if self.img is not None:
            self.img_mod = (transform.rescale(self.img, 1 / factor) * 255).astype(np.uint8)
        else:
            raise ValueError("Image has not been converted from buffer yet!")

    def reduceBuffer(self, factor):
        """Downscales image using SerialEM's implementation of antialiased reduction with a Lanczos 2 filter."""

        sem.ReduceImage(self.buf, factor)
        self.rollBuffers()      # Roll before adding buf_mod to list
        self.buf_mod = "A"

    def addNavPoint(self, point):
        """Adds buffer coords as nav point."""

        # Show buffer
        self.show()

        # Add point
        nav_id = int(sem.AddImagePosAsNavPoint(self.buf, point[1], point[0]))

        return nav_id

    def addNavBoxes(self, boxes, labels=[], padding_factor=1):
        """Adds polygons from boxes in image coords."""

        # Show buffer
        self.show()

        # Find scale factor
        if boxes.pix_size:
            factor = boxes.pix_size / self.pix_size
        else:
            factor = 1

        # Add boxes as polygon
        for b, box in enumerate(boxes.boxes):
            # Scale box coords
            box = box * factor

            # Check for stage limits (currently hardcoded as +/-990, TODO, once SerialEM property becomes readable by script command)
            stage_coords = self.px2stage(box.center) 
            if not (-990 < stage_coords[0] < 990 and -990 < stage_coords[1] < 990):
                continue

            # Pad box
            box.scale(padding_factor)

            # Choose label
            label = ""
            if labels and b < len(labels):
                label = labels[b]
            elif box.label:
                label = box.label

            # Add polygon
            if self.nav is not None:
                # Get z coordinate
                if self.nav_id:
                    stage_z = self.nav.items[self.nav_id].stage[2]
                else:
                    stage_z = 0.0
                # Get xy coordinates
                box_min = self.px2stage(box.xyxycc[:2])
                box_max = self.px2stage(box.xyxycc[2:4])
                pts_x = [box_min[0], box_min[0], box_max[0], box_max[0], box_min[0]]
                pts_y = [box_min[1], box_max[1], box_max[1], box_min[1], box_min[1]]
                # Add nav item
                nav_id = self.nav.newPolygon(pts_x, pts_y, stage_z, label=label, note=f"{label}: {config.WG_model_categories[box.cat]} ({round(box.prob * 100)} %)", color=config.WG_model_nav_colors[box.cat], update=False)
                # Add confidence as UserValue1 for nav item
                self.nav.items[nav_id].addEntry("UserValue1", box.prob)
            else:
                # Add point via SerialEM
                try:
                    nav_id = int(sem.AddImagePosAsNavPoint(self.buf, box.center[0], box.center[1])) - 1
                    # Adjust nav item
                    sem.ChangeItemColor(nav_id + 1, config.WG_model_nav_colors[box.cat])
                    if label:
                        sem.ChangeItemNote(nav_id + 1, f"{label}: {config.WG_model_categories[box.cat]} ({round(box.prob * 100)} %)")
                        sem.ChangeItemLabel(nav_id + 1, label)
                    else:
                        sem.ChangeItemNote(nav_id + 1, f"{config.WG_model_categories[box.cat]} ({round(box.prob * 100)} %)")
                except:
                    log(f"ERROR: Could not create navigator point at {stage_coords}.")

        # Deferred update after adding all boxes
        if self.nav:
            self.nav.push()


    def px2stage(self, point):
        """Converts image coords to stage coords using buffer specific image and montage shifts."""

        return sem.BufImagePosToStagePos(self.buf, 1, point[0], point[1])[:2]
       
    def stage2px(self, point):
        """Converts stage coords to image coords using buffer specific image and montage shifts."""

        return sem.StagePosToBufImagePos(self.buf, 1, point[0], point[1])
    
    def rollBuffers(self, skip=[]):
        """Rolls buffers of all instances of Buffer."""

        # Check all instances
        for buf_ref in self.instances:

            # Get object from weakref object
            buf = buf_ref()

            # Ignore dead refs
            if not buf: continue

            # Move buffer by 1 while in rolling buffers and remove if it moves beyond
            if buf.buf in self.rolling_buffers and buf.buf not in skip:
                buf.buf = chr(ord(buf.buf) + 1)
                if buf.buf not in self.rolling_buffers:
                    buf.buf = None

            # Do the same for the modified buffers
            if buf.buf_mod in self.rolling_buffers and buf.buf_mod not in skip:
                buf.buf_mod = chr(ord(buf.buf_mod) + 1)
                if buf.buf_mod not in self.rolling_buffers:
                    buf.buf_mod = None

    def getCropImage(self, coords, fov):
        """Crops out virtual map at given coordinates and field of view."""

        if self.img is None:
            self.imgFromBuffer()

        log(f"DEBUG:\nMap dimensions: {self.img.shape}\nCrop coords:\nx: {int(coords[0] - fov[0] / 2)}, {int(coords[0] + fov[0] / 2)}\ny: {int(coords[1] - fov[1] / 2)}, {int(coords[1] + fov[1] / 2)}")

        imageCrop = self.img[max(0, int(coords[0] - fov[0] / 2)):min(self.img.shape[0], int(coords[0] + fov[0] / 2)), max(0, int(coords[1] - fov[1] / 2)):min(self.img.shape[1], int(coords[1] + fov[1] / 2))]
        if not np.array_equal(imageCrop.shape, fov):
            mean = np.mean(imageCrop)
            if imageCrop.shape[0] < fov[0]:
                padding = np.full((int(fov[0] - imageCrop.shape[0]), imageCrop.shape[1]), mean)
                if int(coords[0] - fov[0] / 2) < 0:
                    imageCrop = np.concatenate((padding, imageCrop), axis=0)
                if int(coords[0] + fov[0] / 2) > self.img.shape[0]:
                    imageCrop = np.concatenate((imageCrop, padding), axis=0)
            if imageCrop.shape[1] < fov[1]:
                padding = np.full((imageCrop.shape[0], int(fov[1] - imageCrop.shape[1])), mean)
                if int(coords[1] - fov[1] / 2) < 0:
                    imageCrop = np.concatenate((padding, imageCrop), axis=1)
                if int(coords[1] + fov[1] / 2) > self.img.shape[1]:
                    imageCrop = np.concatenate((imageCrop, padding), axis=1)		
            log("WARNING: Target position is close to the edge of the map and was padded.")
        return imageCrop
    
    def __del__(self):
        """Cleans list of instances (but cannot delete current instance)."""

        remaining_refs = []
        for ref in self.__class__.instances:
            if ref():
                remaining_refs.append(ref)
        self.__class__.instances = remaining_refs
    


#!/usr/bin/env python
# ===================================================================
# ScriptName:   scope
# Purpose:      Dummy interface for microscope functions.
#               More information at http://github.com/eisfabian/SPACEtomo
# Author:       Fabian Eisenstein
# Created:      2024/09/02
# Revision:     v1.2
# Last Change:  2024/09/02: separated from modules/scope.py
# ===================================================================

import mrcfile
import numpy as np
from pathlib import Path
from PIL import Image, ImageDraw

import SPACEtomo.config as config
from SPACEtomo.modules.nav import Navigator
from SPACEtomo.modules.dummy.buf import BufferDummy
from SPACEtomo.modules.utils import log

### DEV SETTINGS ###
# 
# Dummy image generation
cats = {
    "background": 0,
    "white": 1,
    "black": 2,
    "crack": 3,
    "coating": 4,
    "cell": 5,
    "cellwall": 6,
    "nucleus": 7,
    "vacuole": 8,
    "mitos": 9,
    "lipiddroplets": 10,
    "vesicles": 11,
    "multivesicles": 12,
    "membranes": 13,
    "dynabeads": 14,
    "ice": 15,
    "cryst": 16
}
cats_intensity = {
    "background": 0.5,
    "white": 0.9,
    "black": 0.1,
    "crack": 0.9,
    "coating": 0.25,
    "cell": 0.4,
    "cellwall": 0.35,
    "nucleus": 0.37,
    "vacuole": 0.32,
    "mitos": 0.3,
    "lipiddroplets": 0.25,
    "vesicles": 0.33,
    "multivesicles": 0.33,
    "membranes": 0.3,
    "dynabeads": 0.2,
    "ice": 0.11,
    "cryst": 0.2
}
#
### END SETTINGS ###

class MicroscopeDummy:

    def __init__(self, imaging_params, cur_dir=None, map_dir=None, nav=None) -> None:

        self.imaging_params = imaging_params
        self.low_dose = False
        self.low_dose_area = None

        # Limits
        self.tilt_limit = 60.0
        self.is_limit = 15
        self.stage_limit = 990

        self.imaging_params.IS_limit = self.is_limit

        # Properties
        self.ta_offset = 0.0

        # Check autoloader
        self.loaded_grid = None
        self.autoloader = {}
        self.checkAutoloader()

        # Dummy state
        if not cur_dir:
            self.cur_dir = Path.cwd()
        else:
            self.cur_dir = Path(cur_dir)
        if not map_dir:
            self.map_dir = Path.cwd() / "SPACE_maps"
        else:
            self.map_dir = Path(map_dir)
        if not nav:
            self.nav = Navigator(Path("temp.nav"))
        else:
            self.nav = nav

        self.dummy_state = {}
        self.dummy_state["stage"] = np.zeros(3)
        self.dummy_state["tilt_angle"] = 0.0
        self.dummy_state["image_shift"] = np.zeros(2)
        self.dummy_state["specimen_shift"] = np.zeros(2)
        self.dummy_state["defocus"] = 0.0
        self.dummy_state["spot_size"] = 6
        self.dummy_state["condenser_lens"] = 2.0
        self.dummy_state["beam_tilt"] = np.zeros(2)
        self.dummy_state["micro_probe"] = 1
        self.dummy_state["magnification"] = 33000
        self.dummy_state["camera"] = [0, None] # camera_id and read_mode
        self.dummy_state["energy_filter"] = [20, 0, 1] # slit_width, energy_loss, slit_in

        if isinstance(self.imaging_params.MM_image_state, list):
            self.dummy_state["imaging_state"] = self.imaging_params.MM_image_state[0]
        else:
            self.dummy_state["imaging_state"] = self.imaging_params.MM_image_state

    ### Properties

    @property
    def stage(self):
        return self.dummy_state["stage"]
    
    @property
    def tilt_angle(self):
        return self.dummy_state["tilt_angle"]
    
    @property
    def image_shift(self):
        return self.dummy_state["image_shift"]
    
    @property
    def speciment_shift(self):
        return self.dummy_state["specimen_shift"]

    @property
    def defocus(self):
        return self.dummy_state["defocus"]
    
    @property
    def spot_size(self):
        return self.dummy_state["spot_size"]
    
    @property
    def condenser_lens(self):
        return self.dummy_state["condenser_lens"]
    
    @property
    def beam_tilt(self):
        return self.dummy_state["beam_tilt"]
    
    @property
    def micro_probe(self):
        return self.dummy_state["micro_probe"]
    
    @property
    def magnification(self):
        return self.dummy_state["magnification"]
    
    @property
    def camera(self):
        camera_id, read_mode = self.dummy_state["camera"]
        return camera_id, read_mode
    
    @property
    def energy_filter(self):
        slit_width, energy_loss, slit_in = self.dummy_state["energy_filter"]
        return slit_in, slit_width, energy_loss
    
    def getMatrices(self, mag_index=0, camera=None):
        matrices = {}
        matrices["s2ss"] = np.array(((-1, 0), (0, -1)))
        matrices["ss2s"] = np.array(((-1, 0), (0, -1)))
        matrices["is2ss"] = np.array(((1, 0), (0, 1)))
        matrices["ss2is"] = np.array(((1, 0), (0, 1)))
        matrices["c2ss"] = np.array(((1, 0), (0, 1)))
        matrices["ss2c"] = np.array(((1, 0), (0, 1)))
        return matrices
    
    ### Setters

    def setImageShift(self, image_shift, relative=False):
        """Applies image shift."""

        if relative:
            self.dummy_state["image_shift"] += image_shift
            log(f"#DUMMY: Image shift was changed by {image_shift} units to {self.dummy_state['image_shift']}!")
        else:
            self.dummy_state["image_shift"] = image_shift
            log(f"#DUMMY: Image shift was changed to {self.dummy_state['image_shift']} units!")

    def setSpecimenShift(self, specimen_shift, relative=False):
        """Applies image shift in specimen coordinates."""
        
        if not relative:
            if self.low_dose_area == "R":
                try:
                    image_shift = self.imaging_params.rec_ss2is_matrix @ specimen_shift
                    self.dummy_state["image_shift"] = image_shift
                except (AttributeError, ValueError):
                    log(f"ERROR: Could not convert specimen shift to image shift.")
                    return
            elif self.low_dose_area == "V":
                try:
                    image_shift = self.imaging_params.view_ss2is_matrix @ specimen_shift
                    self.dummy_state["image_shift"] = image_shift
                except (AttributeError, ValueError):
                    log(f"ERROR: Could not convert specimen shift to image shift.")
                    return
            self.dummy_state["specimen_shift"] = specimen_shift
            log(f"#DUMMY: Image shift was changed to {self.dummy_state['image_shift']} units!")
        else:
            self.dummy_state["specimen_shift"] += specimen_shift
            log(f"#DUMMY: Image shift was changed by {specimen_shift} microns!")

    def setDefocus(self, defocus, relative=False):
        """Applies defocus."""

        if not relative:
            self.dummy_state["defocus"] = defocus
            log(f"#DUMMY: Defocus was set to {defocus} microns!")
        else:
            self.dummy_state["defocus"] += defocus
            log(f"#DUMMY: Defocus was changed by {defocus} microns!")

    def setSpotSize(self, spot_size: int):
        """Sets spot size."""

        self.dummy_state["spot_size"] = spot_size
        log(f"#DUMMY: Spot size was changed to {self.dummy_state['spot_size']}!")

    def setCondenserLens(self, value):
        """Sets condenser lens/illuminated area."""

        self.dummy_state["condenser_lens"] = value
        log(f"#DUMMY: C2/3 was changed to {self.dummy_state['condenser_lens']}!")

    def setBeamTilt(self, tilt_xy):
        """Sets beam tilt."""

        self.dummy_state["beam_tilt"] = np.array(tilt_xy)
        log(f"#DUMMY: Beam tilt was changed to {self.dummy_state['beam_tilt']}!")

    def setMicroProbe(self, probe_mode: int):
        """Turns micro-probe on (1) or off (0)."""

        self.dummy_state["micro_probe"] = probe_mode
        log(f"#DUMMY: Micro-probe was changed to {self.dummy_state['micro_probe']}!")

    def setMag(self, mag: int):
        """Switches mag by number or index."""

        self.dummy_state["magnification"] = mag

        log(f"#DUMMY: Mag was changed to {self.dummy_state['magnification']}!")

    def setCamera(self, camera_id: int, read_mode: int= None):
        """Selects camera."""

        self.dummy_state["camera"] = [camera_id, read_mode]
        log(f"#DUMMY: Camera was changed to {self.dummy_state['camera'][0]} (Read mode: {self.dummy_state['camera'][1]}!")

    def setEnergyFilter(self, slit_in, slit_width, energy_loss=0):
        """Sets energy filter slit settings."""

        self.dummy_state["energy_filter"] = [slit_width, energy_loss, slit_in]
        log(f"#DUMMY: Energy filter slit was changed to {slit_in} (width: {slit_width}, loss: {energy_loss})!")

    def moveStage(self, xyz):
        """Moves stage to x, y, z."""
        
        self.dummy_state["stage"] = np.array(xyz)
        log(f"#DUMMY: Stage was moved to {self.stage}!")

    def tiltStage(self, tilt, backlash=True):
        """Tilts stage to target angle."""

        # Check tilt limit
        if abs(tilt) > self.tilt_limit:
            log(f"ERROR: {tilt} goes beyond maximum stage tilt [{self.tilt_limit}]!")
            return

        # Tilt to angle
        self.dummy_state["tilt_angle"] = tilt
        log(f"#DUMMY: Stage was tilted to {self.tilt_angle}!")

    ### Simple actions

    def record(self, save=""):
        """Records an image and optionally saves it to file."""

        # Figure out what kind of dummy image to generate
        if self.dummy_state["imaging_state"] == self.imaging_params.WG_image_state:
            # Generate LM image
            img_array = self.dummyImageWG()
            pix_size = self.imaging_params.WG_pix_size
        elif self.dummy_state['magnification'] == self.imaging_params.IM_mag_index:
            # Generate IM image
            img_array = self.dummyImageIM()
            pix_size = self.imaging_params.IM_pix_size
        elif self.dummy_state["imaging_state"] == self.imaging_params.MM_image_state or self.dummy_state["imaging_state"] in self.imaging_params.MM_image_state:
            if self.low_dose_area == "V":
                # Generate View image
                pix_size = self.imaging_params.view_pix_size
            elif self.low_dose_area == "R":
                # Generate Record image
                pix_size = self.imaging_params.rec_pix_size
        else:
            # Generate random noise image
            img_array = np.random.rand(*self.imaging_params.cam_dims).astype(np.float32) 
            pix_size = 0.1


        # Find file to save to
        if save:
            save = Path(save)        
            with mrcfile.new(save, overwrite=True) as mrc:
                mrc.set_data(img_array)
                mrc.voxel_size = pix_size * 10 # Angstrom

            BufferDummy.last_dummy_image = save

        log(f"#DUMMY: Image {save} was recorded!")

    def autofocus(self, target=0.0, measure=False):
        """Uses beam tilt routine to determine defocus and optionally adjusts it to target defocus."""
        
        drift = np.random.uniform(0.1, 1)
        if not measure:
            self.dummy_state["defocus"] = target
            measured = self.dummy_state["defocus"]
            log(f"#DUMMY: Focus was adjusted to target: {self.dummy_state['defocus']}")
        else:
            measured = self.dummy_state["defocus"] + np.random.uniform(-1, 1)
            log(f"#DUMMY: Focus was measured: {measured} (drift: {drift}!")

        return measured, drift

    def eucentricity(self, rough=True):
        """Call eucentricity routine."""

        log(f"#DUMMY: Stage was moved to eucentric height!")

    def changeImagingState(self, target_state, low_dose_expected=None):
        """Changes SerialEM imaging state."""

        # Consider state 1 LM state and all others Low Dose states
        if target_state > 1:
            self.low_dose = True
            self.low_dose_area = "R"

        # Deal with list of imaging states for different low dose areas
        if not isinstance(target_state, list):
            target_state = [target_state]

        for state in target_state:
            self.dummy_state["imaging_state"] = state
            log(f"#DUMMY: Changed image state to {state}!")

        # Check if imaging state fits expectation
        if low_dose_expected is not None:
            if self.low_dose != low_dose_expected:
                log(f"WARNING: Imaging state does not fit expected low dose state!")
                return False
        return True

    def changeLowDoseArea(self, area):
        """Changes Low Dose area."""

        if self.low_dose:
            self.low_dose_area = area
            log(f"#DUMMY: Switched Low Dose Area to {self.low_dose_area}!")
        else:
            log("ERROR: Imaging state is not in Low Dose!")

    def changeObjAperture(self, size=0):
        """Inserts objective aperture of given size or retracts aperture if size is 0."""

        log(f"#DUMMY: Inserted objective aperture with size {size}!")

    def openValves(open=True):
        if open:
            # Open column valves
            log(f"#DUMMY: Opened column valves!")
        else:
            # Close column valves
            log(f"#DUMMY: Closed column valves!")

    def checkAutoloader(self):
        """Checks autoloader slots."""

        for slot in range(1, 13):
            self.autoloader[slot] = "G" + str(slot).zfill(2)

        log(f"#DUMMY: Autoloader was checked!")


    def loadGrid(self, grid_slot):
        # Initiate loading
        if grid_slot != self.loaded_grid:
            self.loaded_grid = grid_slot
            log(f"#DUMMY: Loaded grid {self.autoloader[self.loaded_grid]}!")
        else:
            log(f"Grid [{self.autoloader[grid_slot]}] is already loaded.")

    ### Complex actions

    def collectFullMontage(self, model, overlap):
        """Collects full grid montage."""
        
        grid_name = self.autoloader[self.loaded_grid]

        self.changeImagingState(self.imaging_params.WG_image_state)
        if self.low_dose:
            log("WARNING: Whole grid montage image state is in Low Dose! This might result in unreasonably large grid maps.")

        # Figure out appropriate binning to get close to model pix size
        pix_size = model.pix_size
        binning = 1

        # Save pix size
        self.imaging_params.WG_pix_size = pix_size * binning

        # Setup montage
        grid_bar_num = 20
        map_size = 2000 # microns
        lamella_dims = np.array([15, 25]) # width, height in microns

        # Generate grid map
        grid_map = np.full((int(map_size / pix_size * 1000), int(map_size / pix_size * 1000)), 0.25, dtype=np.float32)
        grid_bar_dist = max(grid_map.shape) / grid_bar_num
        grid_bar_width = max(grid_map.shape) / grid_bar_num / 3
        for i in range(grid_bar_num):
            grid_map[:, int(i * grid_bar_dist): int(i * grid_bar_dist + grid_bar_width)] = 0
            grid_map[int(i * grid_bar_dist): int(i * grid_bar_dist + grid_bar_width), :] = 0

        # Generate lamella img
        lamella_width = int(lamella_dims[0] / map_size * max(grid_map.shape))
        lamella_height = int(lamella_dims[1] / map_size * max(grid_map.shape))
        lamella_img = np.full((lamella_height * 2, lamella_width * 2), 0.25, dtype=np.float32)
        # Lamella ice
        lamella_img[int(lamella_img.shape[0] - lamella_height) // 2: int(lamella_img.shape[0] + lamella_height) // 2, int(lamella_img.shape[1] - lamella_width) // 2: int(lamella_img.shape[1] + lamella_width) // 2] = 0.5
        # Lamella trenches
        lamella_img[:int(lamella_img.shape[0] - lamella_height) // 2, int(lamella_img.shape[1] - lamella_width) // 2: int(lamella_img.shape[1] + lamella_width) // 2] = 1
        lamella_img[int(lamella_img.shape[0] + lamella_height) // 2:, int(lamella_img.shape[1] - lamella_width) // 2: int(lamella_img.shape[1] + lamella_width) // 2] = 1

        # Align lamella with tilt axis
        if abs(self.imaging_params.view_ta_rotation) < 45:
            lamella_img = lamella_img.T

        # Add lamellae at random coords
        random_coords = np.random.rand(np.random.randint(1, 5), 2)
        for coords in random_coords:
            x = int(coords[0] * grid_map.shape[0] * 0.8 + 0.1 * grid_map.shape[0])
            y = int(coords[1] * grid_map.shape[1] * 0.8 + 0.1 * grid_map.shape[1])
            grid_map[x: x + lamella_img.shape[0], y: y + lamella_img.shape[1]] = lamella_img

        # Add 25% noise
        noise = np.random.rand(*grid_map.shape)
        grid_map = grid_map * (0.75 + 0.25 * noise.astype(np.float32))

        # Save map
        grid_map_file = self.cur_dir / (grid_name + ".mrc")
        with mrcfile.new(grid_map_file, overwrite=True) as mrc:
            mrc.set_data(grid_map)
            mrc.voxel_size = pix_size * 10

        map_id = self.nav.newMapFromImg(label=grid_name[:6], template_id=None, coords=self.stage[:2], img_file=grid_map_file, note=grid_map_file.name)

        BufferDummy.last_dummy_image = grid_map_file

        log(f"#DUMMY: Collected whole grid montage at [{grid_map_file.stem}]!")

        return map_id

    def collectPolygonMontage(self, poly_id, file, overlap):
        """Collects montage at polygon using View mode."""

        self.changeLowDoseArea("V")
        # Setup polygon montage
        pix_size = self.imaging_params.view_pix_size

        map_size = np.array([35, 40]) / pix_size * 1000       # width, height in pixels
        lamella_dims = np.array([15, 25]) / pix_size * 1000   # width, height in pixels

        # Generate map with black background
        lamella_map = np.full((int(map_size[1]), int(map_size[0])), cats_intensity["black"], dtype=np.float32)
        seg_map = np.full((int(map_size[1]), int(map_size[0])), cats["black"], dtype=np.uint8)

        # Generate lamella
        lamella_width = int(lamella_dims[0])
        lamella_height = int(lamella_dims[1])

        # Lamella background
        lamella_map[int(lamella_map.shape[0] - lamella_height) // 2: int(lamella_map.shape[0] + lamella_height) // 2, int(lamella_map.shape[1] - lamella_width) // 2: int(lamella_map.shape[1] + lamella_width) // 2] = cats_intensity["background"]
        seg_map[int(seg_map.shape[0] - lamella_height) // 2: int(seg_map.shape[0] + lamella_height) // 2, int(seg_map.shape[1] - lamella_width) // 2: int(seg_map.shape[1] + lamella_width) // 2] = cats["background"]

        # Lamella trenches
        lamella_map[:int(lamella_map.shape[0] - lamella_height) // 2, int(lamella_map.shape[1] - lamella_width) // 2: int(lamella_map.shape[1] + lamella_width) // 2] = cats_intensity["white"]
        lamella_map[int(lamella_map.shape[0] + lamella_height) // 2:, int(lamella_map.shape[1] - lamella_width) // 2: int(lamella_map.shape[1] + lamella_width) // 2] = cats_intensity["white"]

        seg_map[:int(seg_map.shape[0] - lamella_height) // 2, int(seg_map.shape[1] - lamella_width) // 2: int(seg_map.shape[1] + lamella_width) // 2] = cats["white"]
        seg_map[int(seg_map.shape[0] + lamella_height) // 2:, int(seg_map.shape[1] - lamella_width) // 2: int(seg_map.shape[1] + lamella_width) // 2] = cats["white"]

        # Make drawing canvas
        lamella_canvas = Image.fromarray(lamella_map)
        lamella_draw = ImageDraw.Draw(lamella_canvas)
        seg_canvas = Image.fromarray(seg_map)
        seg_draw = ImageDraw.Draw(seg_canvas)

        # Draw cell
        cell_size = np.array([5, 7]) / pix_size * 1000
        random_coords = np.random.rand(30, 2) * (lamella_dims - cell_size) + (np.flip(lamella_map.shape) - lamella_dims) // 2
        cell_coords = []
        for c, coords in enumerate(random_coords):
            if min([np.linalg.norm(coords - random) for random in random_coords[c + 1:]] + [1e9]) < min(cell_size): 
                continue
            lamella_draw.ellipse((*coords, *(coords + cell_size)), fill=cats_intensity["cell"], outline=cats_intensity["cellwall"], width=int(0.5 / pix_size * 1000))
            seg_draw.ellipse((*coords, *(coords + cell_size)), fill=cats["cell"], outline=cats["cellwall"], width=int(0.5 / pix_size * 1000))
            cell_coords.append(coords)

        # Draw organelles
        mito_size = np.array([1, 2]) / pix_size * 1000
        nuc_size = np.array([3, 3]) / pix_size * 1000
        vac_size = np.array([3, 3]) / pix_size * 1000

        for cell in cell_coords:
            # Mitos
            random_coords = np.random.rand(3, 2) * (0.8 * cell_size - mito_size) + cell + cell_size * 0.1
            for coords in random_coords:
                lamella_draw.ellipse((*coords, *(coords + mito_size)), fill=cats_intensity["mitos"])
                seg_draw.ellipse((*coords, *(coords + mito_size)), fill=cats["mitos"])

            # Nucleus
            random_coords = np.random.rand(1, 2) * (0.8 * cell_size - nuc_size) + cell + cell_size * 0.1
            for coords in random_coords:
                size = nuc_size * np.random.uniform(0.3, 1)
                lamella_draw.ellipse((*coords, *(coords + size)), fill=cats_intensity["nucleus"])
                seg_draw.ellipse((*coords, *(coords + size)), fill=cats["nucleus"])

            # Vacuole
            random_coords = np.random.rand(1, 2) * (0.8 * cell_size - vac_size) + cell + cell_size * 0.1
            for coords in random_coords:
                size = vac_size * np.random.uniform(0.5, 1)
                lamella_draw.ellipse((*coords, *(coords + size)), fill=cats_intensity["vacuole"])
                seg_draw.ellipse((*coords, *(coords + size)), fill=cats["vacuole"])

        # Draw ice blobs
        max_size = np.array([3, 3]) / pix_size * 1000
        random_coords = np.random.rand(10, 2) * (lamella_dims - max_size) + (np.flip(lamella_map.shape) - lamella_dims) // 2
        for c, coords in enumerate(random_coords):
            size = np.random.uniform() * max_size
            lamella_draw.ellipse((*coords, *(coords + size)), fill=cats_intensity["ice"])
            seg_draw.ellipse((*coords, *(coords + size)), fill=cats["ice"])

        # Revert back to array
        lamella_map = np.array(lamella_canvas)
        seg_map = np.array(seg_canvas)

        # Lamella coating
        seg_map[int((seg_map.shape[0] + lamella_height) // 2 - seg_map.shape[0] / 10): int(seg_map.shape[0] + lamella_height) // 2, int(seg_map.shape[1] - lamella_width) // 2: int(seg_map.shape[1] + lamella_width) // 2] = cats["coating"]
        lamella_map[int((lamella_map.shape[0] + lamella_height) // 2 - lamella_map.shape[0] / 10): int(lamella_map.shape[0] + lamella_height) // 2, int(lamella_map.shape[1] - lamella_width) // 2: int(lamella_map.shape[1] + lamella_width) // 2] = cats_intensity["coating"]

        # Add 50% noise
        noise = np.random.rand(*lamella_map.shape)
        lamella_map = lamella_map * (0.5 + 0.5 * noise.astype(np.float32))

        # Align lamella with tilt axis
        if abs(self.imaging_params.view_ta_rotation) < 45:
            lamella_map = lamella_map.T
            seg_map = seg_map.T

        # Save map
        with mrcfile.new(file, overwrite=True) as mrc:
            mrc.set_data(lamella_map)
            mrc.voxel_size = pix_size * 10

        # Save seg in model pix size
        model_size = np.round(np.flip(seg_map.shape) * pix_size / config.MM_model_pix_size).astype(int)
        Image.fromarray(np.flip(seg_map, axis=0)).resize(tuple(model_size), resample=Image.Resampling.NEAREST).save(self.map_dir / (file.stem + "_seg.png"))

        lamella_name = file.stem.split("_")[-1]
        map_id = self.nav.newMapFromImg(label=lamella_name, template_id=None, coords=self.stage[:2], img_file=file, note=file.name)

        BufferDummy.last_dummy_image = file

        log(f"#DUMMY: Collected medium mag montage at [{file.stem}]!")

        return map_id
    


    # Dummy image generation
    def dummyImageWG(self):
        # Get proper pix size
        pix_size = self.imaging_params.WG_pix_size

        # Define lamella dims
        lamella_dims = np.array([15, 25]) / pix_size * 1000 # width, height in microns

        # Define grid bar distance and width
        grid_bar_dist = 100 / pix_size * 1000 # microns
        grid_bar_width = grid_bar_dist / 3
        grid_bar_num = int(max(self.imaging_params.cam_dims) / grid_bar_dist / pix_size * 1000)

        # Generate map image
        map_img = np.full(self.imaging_params.cam_dims, 0.25, dtype=np.float32)

        # Add grid bars
        for i in range(grid_bar_num):
            map_img[:, int(i * grid_bar_dist): int(i * grid_bar_dist + grid_bar_width)] = 0
            map_img[int(i * grid_bar_dist): int(i * grid_bar_dist + grid_bar_width), :] = 0

        # Generate lamella image
        lamella_width = int(lamella_dims[0])
        lamella_height = int(lamella_dims[1])
        lamella_img = np.full((lamella_height * 2, lamella_width * 2), 0.25, dtype=np.float32)
        # Lamella ice
        lamella_img[int(lamella_img.shape[0] - lamella_height) // 2: int(lamella_img.shape[0] + lamella_height) // 2, int(lamella_img.shape[1] - lamella_width) // 2: int(lamella_img.shape[1] + lamella_width) // 2] = 0.5
        # Lamella trenches
        lamella_img[:int(lamella_img.shape[0] - lamella_height) // 2, int(lamella_img.shape[1] - lamella_width) // 2: int(lamella_img.shape[1] + lamella_width) // 2] = 1
        lamella_img[int(lamella_img.shape[0] + lamella_height) // 2:, int(lamella_img.shape[1] - lamella_width) // 2: int(lamella_img.shape[1] + lamella_width) // 2] = 1

        # Align lamella with tilt axis
        if abs(self.imaging_params.view_ta_rotation) < 45:
            lamella_img = lamella_img.T

        # Add lamellae at random coords
        random_coords = np.random.rand(np.random.randint(1, 5), 2)
        for coords in random_coords:
            x = int(coords[0] * map_img.shape[0] * 0.8 + 0.1 * map_img.shape[0])
            y = int(coords[1] * map_img.shape[1] * 0.8 + 0.1 * map_img.shape[1])
            map_img[x: x + lamella_img.shape[0], y: y + lamella_img.shape[1]] = lamella_img

        # Add 25% noise
        noise = np.random.rand(*map_img.shape)
        map_img = map_img * (0.75 + 0.25 * noise.astype(np.float32))

        log(f"#DUMMY: Generated LM mag image!")
        return map_img

    def dummyImageIM(self):
        # Get proper pix size
        pix_size = self.imaging_params.IM_pix_size

        # Define lamella dims
        lamella_dims = np.array([15, 25]) / pix_size * 1000 # width, height in pixels

        # Generate map image
        map_img = np.full(np.flip(self.imaging_params.cam_dims).astype(int), 0.25, dtype=np.float32)

        # Generate lamella image
        lamella_width = int(lamella_dims[0])
        lamella_height = int(lamella_dims[1])
        lamella_img = np.full((lamella_height * 2, lamella_width * 2), 0.25, dtype=np.float32)
        # Lamella ice
        lamella_img[int(lamella_img.shape[0] - lamella_height) // 2: int(lamella_img.shape[0] + lamella_height) // 2, int(lamella_img.shape[1] - lamella_width) // 2: int(lamella_img.shape[1] + lamella_width) // 2] = 0.5
        # Lamella trenches
        lamella_img[:int(lamella_img.shape[0] - lamella_height) // 2, int(lamella_img.shape[1] - lamella_width) // 2: int(lamella_img.shape[1] + lamella_width) // 2] = 1
        lamella_img[int(lamella_img.shape[0] + lamella_height) // 2:, int(lamella_img.shape[1] - lamella_width) // 2: int(lamella_img.shape[1] + lamella_width) // 2] = 1

        # Align lamella with tilt axis
        if abs(self.imaging_params.view_ta_rotation) < 45:
            lamella_img = lamella_img.T

        # Add lamellae at random coords
        random_coords = np.random.rand(1, 2)
        for coords in random_coords:
            x = int(coords[0] * (map_img.shape[0] - lamella_img.shape[0]))
            y = int(coords[1] * (map_img.shape[1] - lamella_img.shape[1]))
            map_img[x: x + lamella_img.shape[0], y: y + lamella_img.shape[1]] = lamella_img

        # Add 25% noise
        noise = np.random.rand(*map_img.shape)
        map_img = map_img * (0.75 + 0.25 * noise.astype(np.float32))

        log(f"#DUMMY: Generated IM mag image!")
        return map_img

    def dummyImageMM(self):
        pass

    def dummyImageView(self):
        pass

    def dummyImageRec(self):
        pass


#!/usr/bin/env python
# ===================================================================
# ScriptName:   mod_wg
# Purpose:      Classes for lamella detection.
#               More information at http://github.com/eisfabian/SPACEtomo
# Author:       Fabian Eisenstein
# Created:      2024/08/14
# Revision:     v1.2
# Last Change:  2024/08/21: minor fixes after external test run
#               2024/08/16: added sorting function, added meta data
# ===================================================================

import time
import json
from pathlib import Path
import numpy as np
from PIL import Image
Image.MAX_IMAGE_PIXELS = None           # removes limit for large images

from SPACEtomo.modules import utils
from SPACEtomo.modules.utils import log
import SPACEtomo.config as config

class WGModel:
    def __init__(self, external=False):
        if external:
            self.model = None
        else:
            from ultralytics import YOLO, settings
            model_file = Path(config.WG_model_file)
            if not model_file.exists():
                raise FileNotFoundError("Lamella detection model file does not exist! Please import a model using the SPACEmodel command!")
            self.model = YOLO(Path(config.WG_model_file))
            settings.update({'sync': False})    # no tracking by google analytics

        # Get params from config file
        self.pix_size = config.WG_model_pix_size
        self.sidelen = config.WG_model_sidelen
        self.categories = config.WG_model_categories
        self.cat_colors = np.array(config.WG_model_gui_colors)
        for c in range(len(self.cat_colors)):
            if not isinstance(self.cat_colors[c], str) and any([val > 1 for val in self.cat_colors[c]]): 
                self.cat_colors[c] = self.cat_colors[c] / 255        # convert colors to float if given as RGB
        self.cat_nav_colors = config.WG_model_nav_colors

    def findLamellae(self, map_dir, map_name, suffix="_wg.png", overlap_microns=10, threshold=0, save_boxes=True, device="cpu"):
        """Runs YOLO lamella detection on image file and compiles boxes."""

        # Load image from file
        map_dir = Path(map_dir)
        image = np.array(Image.open(map_dir / (map_name + suffix)))

        # Calculate overlap of tiles in pixels
        overlap = overlap_microns * 1000 // self.pix_size

        # Split up image if larger than model input
        if image.shape[0] > self.sidelen or image.shape[1] > self.sidelen:
            num_cols = -int(image.shape[0] // -(self.sidelen - overlap))      # Using -(a // -b) instead of a // b for ceil int division
            num_rows = -int(image.shape[1] // -(self.sidelen - overlap))
        else:       # avoid additional columns consisting just of overlap
            num_cols = num_rows = 1

        # Pad if smaller than model input
        if image.shape[0] < self.sidelen or image.shape[1] < self.sidelen:
            img_pad = np.zeros((self.sidelen, self.sidelen))
            img_pad[(self.sidelen - image.shape[0]) // 2: (self.sidelen + image.shape[0]) // 2, (self.sidelen - image.shape[1]) // 2: (self.sidelen + image.shape[1]) // 2] = image
            padding = np.array([self.sidelen - image.shape[0], self.sidelen - image.shape[1]])
        else:
            img_pad = image
            padding = np.zeros(2)

        bboxes = []
        for c in range(num_cols):
            for r in range(num_rows):
                x = int(c * (self.sidelen - overlap))
                y = int(r * (self.sidelen - overlap))

                crop = img_pad[x: x + self.sidelen, y: y + self.sidelen]
                crop = np.dstack([crop, crop, crop])

                results = self.model.predict(crop, device=device)                                         # YOLO inference
                
                if len(results[0].boxes) > 0:
                    bbox = np.array(results[0].boxes.xyxy.to("cpu"))

                    # Subtract tile coords and/or padding from coords and make sure coords go not out of image bounds
                    bbox[:, 0] = np.clip(bbox[:, 0] + y - padding[1] // 2, 0, image.shape[1])
                    bbox[:, 2] = np.clip(bbox[:, 2] + y - padding[1] // 2, 0, image.shape[1])
                    bbox[:, 1] = np.clip(bbox[:, 1] + x - padding[0] // 2, 0, image.shape[0])
                    bbox[:, 3] = np.clip(bbox[:, 3] + x - padding[0] // 2, 0, image.shape[0])
                    
                    cat = np.reshape(results[0].boxes.cls.to("cpu"), (bbox.shape[0], 1))
                    conf = np.reshape(results[0].boxes.conf.to("cpu"), (bbox.shape[0], 1))

                    bbox = np.hstack([bbox, cat, conf])

                    if len(bboxes) > 0:
                        bboxes = np.concatenate([bboxes, bbox])
                    else:
                        bboxes = bbox
                    
        bboxes = Boxes(bboxes, label_prefix="PL", pix_size=config.WG_model_pix_size, img_size=image.shape)
        log(f"Lamellae found (initial): \t{len(bboxes)}")

        if len(bboxes) > 0:
            # Merge bboxes found on different crop windows
            bboxes.mergeOverlappingBoxes()
            #bboxes = self.mergeOverlappingBoxes(bboxes, overlap)
            log(f"Lamellae found (merged): \t{len(bboxes)}")

        # Clean lamellae based on user defined confidence threshold
        #bboxes = np.array([bbox for bbox in bboxes if bbox[5] >= threshold])
        bboxes.removeLowConfidence(threshold)
        log(f"Lamellae found (final): \t{len(bboxes)}")

        # Save bboxes to output file
        if save_boxes:
            box_file = map_dir / (Path(map_name + suffix).stem + "_boxes.json")
            bboxes.saveFile(box_file)

        return bboxes

class Boxes:
    """Collection of bounding boxes"""

    def __init__(self, bboxes, label_prefix="", exclude_cats=[], pix_size=None, img_size=None) -> None:
        """Initializes from box list or from box file."""

        self.pix_size = float(pix_size) if pix_size is not None else None       # nm
        self.img_size = np.array(img_size) if img_size is not None else None

        if isinstance(bboxes, Path):
            bboxes = self.getFromFile(bboxes)

        self.boxes = []
        self.excluded_boxes = []
        for b, box in enumerate(bboxes):
            # Ignore boxes of exclude_cats
            if config.WG_model_categories[int(box[4])] in exclude_cats:
                self.excluded_boxes.append(BBox(box[:4], *box[5:], label=label_prefix + str(b + 1)))
            else:
                self.boxes.append(BBox(box[:4], *box[4:], label=label_prefix + str(b + 1)))

        if len(self.boxes) - len(bboxes) > 0:
            log(f"NOTE: {len(self.boxes) - len(bboxes)} lamellae were excluded because they were {', '.join(exclude_cats)}!")
    
    def mergeOverlappingBoxes(self, margin=0.1):
        """Merges overlapping bboxes."""

        # Sort boxes by x values
        self.sortBy("x")

        merged_boxes = []
        for box in self.boxes:
            if merged_boxes:
                # Get the previous merged box
                prev_box = merged_boxes[-1]

                #size = [box[2] - box[0], box[3] - box[1]]

                # Check if the current box overlaps with the previous box within margin
                if box.x_min <= prev_box.x_max + margin * box.size[0] and box.y_min <= prev_box.y_max + margin * box.size[1] and box.y_max >= prev_box.y_min - margin * box.size[1]:
                    # Update the previous box to encompass both boxes
                    prev_box.x_max = max(box.x_max, prev_box.x_max)
                    prev_box.y_max = max(box.y_max, prev_box.y_max)
                    prev_box.y_min = min(box.y_min, prev_box.y_min)
                    # Change class if probability is higher for new box OR if old box was "undesirable" class
                    if prev_box.prob < box.prob or config.WG_model_categories[prev_box.cat] == "broken" or config.WG_model_categories[prev_box.cat] == "gone":
                        prev_box.cat = box.cat
                        prev_box.prob = box.prob
                else:
                    merged_boxes.append(box)
            else:
                merged_boxes.append(box)
        
        self.boxes = merged_boxes

    def removeLowConfidence(self, threshold):
        """Removes boxes below confidence threshold."""

        new_boxes = []
        for box in self.boxes:
            if box.prob > threshold:
                new_boxes.append(box)
        self.boxes = new_boxes

    def saveFile(self, file):
        """Saves json file with bounding boxes."""

        bboxes = []
        for box in self.boxes:
            bboxes.append(box.xyxycc.tolist())
        
        meta_data = {"pix_size": self.pix_size, "img_size": self.img_size, "datetime": time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())}

        data = {"boxes": bboxes, "meta_data": meta_data}
        with open(file, "w") as f:
            json.dump(data, f, indent=4, default=utils.convertArray)

    def getFromFile(self, file):
        """Reads json box file."""

        with open(file, "r") as f:
            data = json.load(f, object_hook=utils.revertArray)
            if isinstance(data, dict):
                bboxes = data["boxes"]
                meta_data = data["meta_data"]
            else:
                bboxes = data           # keep compatibility with boxes.json before meta_data

        if "pix_size" in meta_data.keys():
            self.pix_size = meta_data["pix_size"]
        if "img_size" in meta_data.keys():
            self.img_size = meta_data["img_size"]

        return bboxes
    
    def sortBy(self, key, reverse=False):
        """Sorts boxes by coords, distance or attribute."""

        if len(self) <= 1: return

        if key == "x":
            if not reverse:
                self.boxes = sorted(self.boxes, key=lambda box: box.x_min)
            else:
                self.boxes = sorted(self.boxes, key=lambda box: box.x_max, reverse=True)
        elif key == "y":
            if not reverse:
                self.boxes = sorted(self.boxes, key=lambda box: box.y_min)
            else:
                self.boxes = sorted(self.boxes, key=lambda box: box.y_max, reverse=True)
        elif key == "center_dist":
            if self.img_size is not None:
                self.boxes = sorted(self.boxes, key=lambda box: np.linalg.norm(box.center - self.img_size // 2), reverse=reverse)
            else:
                log("ERROR: Can't sort by distance from center, because image size is not known!")
        elif hasattr(self.boxes[0], key):
            self.boxes = sorted(self.boxes, key=lambda box: getattr(box, key), reverse=reverse)
        else:
            raise NotImplementedError("Sorting key not available!")

    def __bool__(self):
        return bool(self.boxes)

    def __len__(self):
        return len(self.boxes)
    

class BBox:
    """Class for bounding box."""

    def __init__(self, bbox, cat, prob=1, label="") -> None:
        self.x_min = bbox[0]
        self.x_max = bbox[2]
        self.y_min = bbox[1]
        self.y_max = bbox[3]
        self.cat = int(cat)
        self.prob = prob

        self.label = label

    @property
    def center(self):
        return np.array([(self.x_min + self.x_max) / 2, (self.y_min + self.y_max) / 2])
    
    @property
    def size(self):
        return np.array([self.x_max - self.x_min, self.y_max - self.y_min])
    
    @property
    def yolo(self):
        """Box in format CXYWH (category, center x, center y, width, height)"""
        return np.array([self.cat, *self.center, *self.size])
    
    @property
    def xyxycc(self):
        """Box in format XYXYCC (min x, min y, max x, max y, category, confidence)"""
        return np.array([self.x_min, self.y_min, self.x_max, self.y_max, self.cat, self.prob])
    
    def scale(self, factor):
        size = self.size * factor
        self.x_min, self.y_min = self.center - size / 2
        self.x_max, self.y_max = self.center + size / 2
       
    def __mul__(self, mult):
        bbox = self.xyxycc[:4] * mult
        return BBox(bbox, self.cat, self.prob, self.label)
    
    __rmul__ = __mul__
        